#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org

import logging

import math
import numpy as np
import scipy.constants as unit
from scipy.integrate import odeint

from cortix import Module
from cortix.support.phase_new import PhaseNew as Phase
from cortix import Quantity

class UMLRR(Module):
    '''
    UMass Lowell research nuclear reactor single-point model.

    Notes
    -----
    These are the `port` names available in this module

      coolant-inflow, coolant-outflow, signal-out, signal-in

    to connect to other modules if any.
    See instance attribute `port_names_expected`.

    '''

    def __init__(self):
        '''
        Parameters
        ----------
        params: dict
            All parameters for the module in the form of a dictionary.

        '''

        super().__init__()

        self.port_names_expected = ['coolant-inflow', 'coolant-outflow',
                                    'signal-out', 'signal-in']

        unit.kg     = unit.kilo*unit.gram
        unit.meter  = 1.0
        unit.second = 1.0
        unit.pascal = 1.0
        unit.joule  = 1.0
        unit.kj     = unit.kilo*unit.joule
        unit.kelvin = 1.0
        unit.watt   = 1.0
        unit.barn   = 1.0e-28 * unit.meter**2

        self.initial_time = 0.0 * unit.day
        self.end_time     = 4 * unit.hour
        self.time_step    = 10.0 * unit.second
        self.show_time    = (False,10.0)

        self.log = logging.getLogger('cortix')

        self.params = dict()

        # General parameters

        #self.params['gen-time'] = 1.0e-4  # s
        self.params['gen-time'] = 6.5e-5  # s
        #self.params['beta']     = 6.5e-3  # 
        self.params['beta']     = 7.8e-3  # 
        #self.params['species-decay']     = [0.0124, 0.0305, 0.111, 0.301, 1.14, 3.01] # 1/sec
        self.params['species-decay']     = [0.0127, 0.0317, 0.1160, 0.3111, 1.4003, 3.8708] # 1/sec
        #self.params['species_rel_yield'] = [0.033, 0.219, 0.196, 0.395, 0.115, 0.042]
        self.params['xi'] = [0.00026, 0.00146, 0.00129, 0.00279, 0.0008, 0.00018]
        beta_i = xi*beta/sum(xi)

        self.params['alpha_n']       = -5e-4 # control rod reactivity worth
        self.params['alpha_n']       = -5e-5 # control rod reactivity worth
        self.params['alpha_tn_fake'] = -1e-4/20 # -1.0e-6

        self.params['n_dens_ss_operation'] = 1e15 * 1e4 / 2200  # neutrons/m^2

        self.params['fis_energy']           = 180 * 1.602e-13 * unit.joule # per fission 
        self.params['sigma_f_o']            = 586.2 * unit.barn
        self.params['temp_o']               = unit.convert_temperature(20,'C','K')
        self.params['temp_c_ss_operation']  = unit.convert_temperature(550,'C','K') # desired ss operation temp of coolant
        self.params['temp_f_safe_max']      = unit.convert_temperature(1100,'C','K')
        self.params['thermal_neutron_velo'] = 2200 * unit.meter/unit.second
        self.params['fis_nuclide_num_dens_fake'] = 1e17/40 * 1.0e+6 # (fissile nuclei)/m3

        self.params['fuel_dens']   = 2500 * unit.kg/unit.meter**3
        self.params['cp_fuel']     = 720  * unit.joule/unit.kg/unit.kelvin
        #self.params['fuel_volume'] = 1.5  * unit.meter**3
        self.params['fuel_volume'] = 1.5  * unit.meter**3

        self.params['coolant_dens']   = 0.1786 * unit.kg/unit.meter**3
        self.params['cp_coolant']     = 20.78/4e-3 * unit.joule/unit.kg/unit.kelvin
        self.params['coolant_volume'] = 0.8 * unit.meter**3

        self.params['ht_coeff'] = 800 * unit.watt/unit.kelvin

        self.params['strict'] = True # apply strict testing to some quantities

        self.params['shutdown']      = False
        self.params['shutdown_time'] = 0.0 # s
        self.params['rho_shutdown']  = 0.0 # s

        self.params['coolant_flowrate_forced'] = 1650.0 * unit.gallon/unit.minute

        # Initial data parameters

        gen_time = self.params['gen-time'] # retrieve neutron generation time
        self.params['q_0'] = 1/gen_time # pulse at t = 0
        self.params['n_ss']       = 0.0 # neutronless steady state before start up
        self.params['n_dens_ref'] = 1.0
        rho_0_over_beta = 0.5 # $
        beta = self.params['beta'] # retrieve the delayed neutron fraction
        self.params['reactivity'] = rho_0_over_beta * beta # "rho/beta = 10 cents"
        self.params['temp_0'] = self.params['temp_o']
        self.params['tau_fake'] = .025 # s residence time

        # Setup steady state initial conditions

        n_species = len(self.params['species-decay'])

        assert len(self.params['species_rel_yield']) == n_species

        c_vec_0 = np.zeros(n_species,dtype=np.float64) # initialize conentration vector

        species_decay = self.params['species-decay'] # retrieve list of decay constants
        lambda_vec    = np.array(species_decay) # create a numpy vector

        species_rel_yield = self.params['species_rel_yield']
        beta = self.params['beta']
        beta_vec = np.array(species_rel_yield) * beta  # create the beta_i's vector

        gen_time = self.params['gen-time'] # retrieve neutron generation time

        n_ss = self.params['n_ss']
        c_vec_ss = beta_vec/lambda_vec/gen_time * n_ss # compute the steady state precursors number density

        self.params['c_vec_ss'] = c_vec_ss

        # setup initial condition for variables
        self.params['n_0']     = n_ss
        self.params['c_vec_0'] = c_vec_ss
        self.params['rho_0']   = self.params['reactivity']

        self.params['temp_f_0'] = self.params['temp_0'] + 10.0 # helps startup integration
        self.params['temp_c_0'] = self.params['temp_0']

        # Reactor phase history
        quantities = list()

        neutron_dens = Quantity( name='neutron-dens', formal_name='n', unit='1/m^3',
                value=self.params['n_0'],
                info='Reactor Neutron Density', latex_name='$n$')

        quantities.append(neutron_dens)

        delayed_neutron_cc = Quantity( name='delayed-neutrons-cc', formal_name='c_i',
                unit='1/m^3', value= self.params['c_vec_0'],
                info='Delayed Neutron Precursors', latex_name='$c_i$')

        quantities.append(delayed_neutron_cc)

        fuel_temp = Quantity( name='fuel-temp', formal_name='T_f', unit='K',
                value=self.params['temp_f_0'], info='Reactor Nuclear Fuel Temperature',
                latex_name='$T_f$')

        quantities.append(fuel_temp)

        temp_in = self.params['temp_c_0']
        pwr_0 = self.params['coolant_volume']/self.params['tau_fake'] * \
                self.params['coolant_dens'] * self.params['cp_coolant'] * \
                (self.params['temp_c_0'] - temp_in)

        power = Quantity( name='power', formal_name='Pwr', unit='W',
                value=pwr_0, info='Reactor Power',
                latex_name='$\mathrm{Prw}$')

        quantities.append(power)

        self.reactor_phase = Phase(self.initial_time, time_unit='s',
                quantities=quantities)

        # Coolant outflow phase history
        quantities = list()

        flowrate = Quantity( name='flowrate', formal_name='q_c', unit='kg/s', value=0.0,
                info='Reactor Outflow Coolant Flowrate', latex_name='$q_c$')

        quantities.append(flowrate)

        temp = Quantity( name='temp', formal_name='T_c', unit='K',
                value=self.params['temp_c_0'],
                info='Reactor Outflow Coolant Temperature', latex_name='$T_c$')

        quantities.append(temp)

        self.coolant_phase = Phase(self.initial_time, time_unit='s',
                quantities=quantities)

        # Initialize inflow
        self.params['inflow-cool-temp'] = self.params['temp_c_0']

        self.__logit = False # flag for when to log throughout the code

        return

    def run(self):

        # Some logic for logging time stamps
        if self.initial_time + self.time_step > self.end_time:
            self.end_time = self.initial_time + self.time_step

        time = self.initial_time

        print_time = self.initial_time
        print_time_step = self.show_time[1]
        if print_time_step < self.time_step:
            print_time_step = self.time_step

        while time <= self.end_time:

            last_time_stamp = time - self.time_step

            if self.show_time[0] and time>=print_time and \
                    time<print_time+print_time_step:

                self.log.info(self.name+'::run():time[m]='+ str(round(time/unit.minute,1)))

                self.__logit = True
                print_time += self.show_time[1]

            else:
                self.__logit = False

            # Communicate information
            #------------------------
            self.__call_ports(time)

            # Evolve one time step
            #---------------------

            time = self.__step( time )

    def __call_ports(self, time):

        # Interactions in the coolant-inflow port
        #----------------------------------------
        # one way "from" coolant-inflow

        # receive from
        if self.get_port('coolant-inflow').is_connected:

            self.send( time, 'coolant-inflow' )
            (check_time, coolant_stream) = self.recv('coolant-inflow')
            assert abs(check_time-time) <= 1e-6
            self.params['inflow-cool-temp'] = coolant_stream['temp']
        #
        # removed this later; must have a heat exchanger
        # 
        else:
            coolant_stream = self.__get_coolant_stream( time )
            if coolant_stream['temp'] < self.params['temp_c_ss_operation']:
                self.params['inflow-cool-temp'] = coolant_stream['temp']
            else:
                self.params['inflow-cool-temp'] = self.params['temp_c_ss_operation']

        # Interactions in the coolant-outflow port
        #-----------------------------------------
        # one way "to" coolant-outflow

        # send to 
        if self.get_port('coolant-outflow').is_connected:

            message_time = self.recv('coolant-outflow')
            coolant_stream = self.__get_coolant_stream( message_time )
            self.send( (message_time, coolant_stream), 'coolant-outflow' )

        # Interactions in the signal-out port
        #-----------------------------------------
        # one way "to" signal-out 

        # send to 
        if self.get_port('signal-out').is_connected:

            message_time = self.recv('signal-out')
            signal_out = self.__get_signal_out(time)
            self.send( (message_time, signal_out), 'signal-out' )

        # Interactions in the signal-in port
        #-----------------------------------------
        # one way "from" signal-in 

        # receive from
        if self.get_port('signal-in').is_connected:

            self.send( time, 'signal-in' )
            (check_time, signal_in) = self.recv('signal-in')
            assert abs(check_time-time) <= 1e-6
            self.params['signal-in'] = signal_in

    def __step(self, time=0.0):
        r'''
        ODE IVP problem:
        Given the initial data at :math:`t=0`,
        :math:`u = (u_1(0),u_2(0),\ldots)`
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None

        '''

        # Get state values
        u_0 = self.__get_state_vector( time )

        t_interval = np.linspace(time, time+self.time_step, num=2)

        (u_vec_hist, info_dict) = odeint( self.__f_vec,
                                          u_0, t_interval,
                                          args=( self.params, ),
                                          rtol=1e-4, atol=1e-8, mxstep=200,
                                          full_output=True )

        assert info_dict['message'] =='Integration successful.', info_dict['message']

        u_vec = u_vec_hist[1,:]  # solution vector at final time step

        n_dens    = u_vec[0]
        c_vec     = u_vec[1:7]
        fuel_temp = u_vec[7]
        cool_temp = u_vec[8]

        outflow = self.coolant_phase.get_row(time)
        reactor = self.reactor_phase.get_row(time)

        # Advance time
        time += self.time_step

        # Update state variables
        self.reactor_phase.add_row(time, reactor)
        self.coolant_phase.add_row(time, outflow)

        self.reactor_phase.set_value('neutron-dens', n_dens, time)
        self.reactor_phase.set_value('delayed-neutrons-cc', c_vec, time)
        self.reactor_phase.set_value('fuel-temp', fuel_temp, time)

        temp_in = self.params['inflow-cool-temp']
        pwr = self.params['coolant_volume']/self.params['tau_fake'] * \
              self.params['coolant_dens'] * self.params['cp_coolant'] * \
              (cool_temp - temp_in)
        self.reactor_phase.set_value('power', pwr, time)

        self.coolant_phase.set_value('temp', cool_temp, time)

        return time

    def __signal_out(self, time=0.0):
        '''
        Send reactor signals to anyone requesting it.

        Parameters
        ----------
        time: float
            Time.

        Returns
        -------
        None

        '''

        signals = dict()
        signals['time'] = time

        return signals

    def __get_coolant_stream(self, time=0.0):
        '''
        Get the coolant outflow stream. Fully mixed system.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None

        '''

        coolant_stream = dict()

        outflow_cool_temp = self.coolant_phase.get_value('temp', time)

        coolant_stream['temp'] = outflow_cool_temp

        return coolant_stream

    def __get_state_vector(self, time):
        '''
        Return a numpy array of all unknowns ordered as below:
        neutron density (1), delayed neutron emmiter concentrations (6),
        termperature of fuel (1), temperature of coolant (1).
        '''

        u_vec = np.empty(0,dtype=np.float64)

        neutron_dens = self.reactor_phase.get_value('neutron-dens',time)
        u_vec = np.append( u_vec, neutron_dens )

        delayed_neutrons_cc = self.reactor_phase.get_value('delayed-neutrons-cc',time)
        u_vec = np.append(u_vec, delayed_neutrons_cc)

        fuel_temp = self.reactor_phase.get_value('fuel-temp',time)
        u_vec = np.append( u_vec, fuel_temp)

        temp = self.coolant_phase.get_value('temp',time)
        u_vec = np.append(u_vec, temp)

        # sanity check
        assert not u_vec[u_vec<0.0],'%r'%u_vec

        return u_vec

    def __alpha_tn_func(self, temp, params):
        '''
        Reactivity temperature feedback coefficient.

        '''

        alpha_tn = params['alpha_tn_fake']

        return alpha_tn

    def __rho_func(self, t, n_dens, temp, params):
        '''
        Reactivity function.

        Parameters
        ----------
        t: float, required
            Time.
        temp_f: float, required
            Temperature at time t.
        params: dict, required
            Dictionary of quantities. It must have a `'rho_0'` key/value pair.

        Returns
        -------
        rho_t: float
            Value of reactivity.

        Examples
        --------

        '''

        if params['shutdown'] == False or \
           (params['shutdown'] == True and time < params['shutdown_time']):

            rho_0 = params['rho_0']
            n_dens_ref = params['n_dens_ref']
            temp_ref = params['temp_c_ss_operation']
            alpha_n  = params['alpha_n']
            alpha_tn = self.__alpha_tn_func( temp, params )

        elif params['shutdown'] == True and time >= params['shutdown_time']:

            rho_0 = params['rho_shutdown']
            n_dens_ref = 0.0
            temp_ref = params['temp_o']
            alpha_n  = 0
            alpha_tn = 0

        else:

            assert False

        rho_t = rho_0 + alpha_n * (n_dens - n_dens_ref) + alpha_tn * (temp - temp_ref)

        beta = params['beta']

        if params['strict'] == True:
            assert rho_t/beta < 1,\
            'rho/beta = %r at time = %r; rho_0 = %r; rho_n = %r; rho_tn = %r'%\
            ( rho_t/beta, time, rho_0/beta, alpha_n*(n_dens - n_dens_ref)/beta, \
              alpha_tn*(temp - temp_ref)/beta )

        return rho_t

    def __q_source(self, t, params):
        '''
        Neutron source delta function.

        Parameters
        ----------
        t: float, required
            Time.
        params: dict, required
            Dictionary of quantities. It must have a `'q_0'` key/value pair.

        Returns
        -------
        q: float
            Value of source.

        Examples
        --------

        '''

        q = 0.0
        q_0 = params['q_0']

        if t <= 1e-5: # small time value
            q = q_0
        else:
            q = 0.0

        return q

    def __sigma_fis_func(self, temp, params):
        '''
        Place holder for implementation
        '''

        sigma_f = params['sigma_f_o']

        return sigma_f

    def __nuclear_pwr_dens_func(self, time, temp, n_dens, params ):
        '''
        Place holder for implementation
        '''

        rxn_heat = params['fis_energy'] # get fission reaction energy J per reaction

        sigma_f = self.__sigma_fis_func( temp, params ) # m2

        fis_nuclide_num_dens = params['fis_nuclide_num_dens_fake'] #  #/m3

        Sigma_fis = sigma_f * fis_nuclide_num_dens # macroscopic cross section

        v_o = params['thermal_neutron_velo'] # m/s

        neutron_flux = n_dens * params['n_dens_ss_operation'] * v_o

        # reaction rate density
        rxn_rate_dens = Sigma_fis * neutron_flux

        # nuclear power source
        q3prime = - rxn_heat * rxn_rate_dens # exothermic reaction W/m3

        if params['strict'] == True:
            assert q3prime <= 0.0,"time = %r, q''' = %r, n_dens = %r"%(time,q3prime,n_dens)

        return q3prime

    def __heat_sink_rate(self, time, temp_f, temp_c, params):

        ht_coeff = params['ht_coeff']

        q_f = - ht_coeff * (temp_f - temp_c)

        if params['strict'] == True:
            assert q_f <= 0.0,'q_f = %r at time = %r; temp_f = %r, temp_c = %r'%(q_f,time,temp_f,temp_c)

        return q_f

    def __f_vec(self, u_vec, time, params):
        '''
        ODE RHS function
        '''

        import numpy as np
        #assert np.all(u_vec >= 0.0),'time = %r; u_vec = %r'%(time,u_vec)

        n_dens = u_vec[0] # get neutron dens

        c_vec = u_vec[1:-2] # get delayed neutron emitter concentration

        temp_f = u_vec[-2] # get temperature of fuel

        temp_c = u_vec[-1] # get temperature of coolant

        # initialize f_vec to zero
        species_decay = params['species-decay']
        lambda_vec    = np.array(species_decay)
        n_species     = len(lambda_vec)

        assert len(lambda_vec)==len(c_vec)

        f_tmp = np.zeros(1+n_species+2,dtype=np.float64) # vector for f_vec return

        #----------------
        # neutron balance
        #----------------
        rho_t    = self.__rho_func(time, n_dens, (temp_f+temp_c)/2.0, params)

        beta     = params['beta']
        gen_time = params['gen-time']

        q_source_t = self.__q_source(time, params)

        f_tmp[0] = (rho_t - beta)/gen_time * n_dens + lambda_vec @ c_vec + q_source_t

        #-----------------------------------
        # n species balances (implicit loop)
        #-----------------------------------

        species_rel_yield = params['species_rel_yield']
        beta_vec = np.array(species_rel_yield) * beta

        assert len(beta_vec)==len(lambda_vec)

        f_tmp[1:-2] = beta_vec/gen_time * n_dens - lambda_vec * c_vec

        #--------------------
        # fuel energy balance
        #--------------------
        rho_f    = params['fuel_dens']
        cp_f     = params['cp_fuel']
        vol_fuel = params['fuel_volume']

        pwr_dens = self.__nuclear_pwr_dens_func( time, (temp_f+temp_c)/2, n_dens, params )

        heat_sink = self.__heat_sink_rate( time, temp_f, temp_c, params )

        f_tmp[-2] = - 1/rho_f/cp_f * ( pwr_dens - heat_sink/vol_fuel )

        #-----------------------
        # coolant energy balance
        #-----------------------
        rho_c    = params['coolant_dens']
        cp_c     = params['cp_coolant']
        vol_cool = params['coolant_volume']

        temp_in = params['inflow-cool-temp']

        tau = params['tau_fake']

        heat_source = - heat_sink

        f_tmp[-1] = - 1/tau * (temp_c - temp_in) + 1./rho_c/cp_c/vol_cool * heat_source

        return f_tmp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org

import numpy as np
import scipy.constants as unit
from scipy.integrate import odeint

from cortix import Module
from cortix import Phase
from cortix import Quantity

class Probation(Module):
    '''
    Probation Cortix module used to model criminal group population in a probation.

    Notes
    -----
    These are the `port` names available in this module to connect to respective
    modules: `adjudication`, `jail`, `arrested`, and `community`.
    See instance attribute `port_names_expected`.

    '''

    def __init__(self, n_groups=1, pool_size=0.0):
        '''
        Parameters
        ----------
        n_groups: int
            Number of groups in the population.
        pool_size: float
            Upperbound on the range of the existing population groups. A random value
            from 0 to the upperbound value will be assigned to each group.

        '''

        super().__init__()

        self.port_names_expected = ['adjudication','jail','arrested','community']

        quantities      = list()
        self.ode_params = dict()

        self.initial_time = 0.0 * unit.day
        self.end_time     = 100 * unit.day
        self.time_step    = 0.5 * unit.day

        unit.month = 30*unit.day
        unit.percent = 1/100

        # Population groups
        self.n_groups = n_groups

        # Probation population groups
        fbg_0 = np.random.random(self.n_groups) * pool_size
        fbg = Quantity(name='fbg', formal_name='probation-pop-grps',
                latex_name = '$n_b^{(g)}$',
                unit='# offenders', value=fbg_0, info='Probation Population Groups')
        quantities.append(fbg)

        # Model parameters: commitment coefficients

        # Probation to community
        a = 35*unit.percent/unit.year * np.ones(self.n_groups)
        b = 45*unit.percent/unit.year * np.ones(self.n_groups)
        cb0g_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        cb0g = Quantity(name='cb0g', formal_name='commit-community-coeff-grps',
               unit='1/s', value=cb0g_0)
        self.ode_params['commit-to-community-coeff-grps'] = cb0g_0
        quantities.append(cb0g)

        # Probation to jail
        a = 55*unit.percent/unit.year * np.ones(self.n_groups)
        b = 65*unit.percent/unit.year * np.ones(self.n_groups)
        cbjg_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        cbjg = Quantity(name='cbjg', formal_name='commit-jail-coeff-grps',
               unit='1/s', value=cbjg_0)
        self.ode_params['commit-to-jail-coeff-grps'] = cbjg_0
        quantities.append(cbjg)

        # Death term
        a = 0.5*unit.percent/unit.year * np.ones(self.n_groups)
        b = 0.7*unit.percent/unit.year * np.ones(self.n_groups)
        dbg_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        self.ode_params['probation-death-rates-coeff'] = dbg_0

        # Initialize inflows to zero
        self.ode_params['arrested-inflow-rates']     = np.zeros(self.n_groups)
        self.ode_params['adjudication-inflow-rates'] = np.zeros(self.n_groups)

        # Phase state
        self.population_phase = Phase(self.initial_time, time_unit='s',
                quantities=quantities)

        self.population_phase.SetValue('fbg', fbg_0, self.initial_time)

        return

    def run(self, *args):

        self.__zero_ode_parameters()

        time = self.initial_time

        while time <= self.end_time:

            # Interactions in the jail port
            #------------------------------
            # one way "to" jail

            message_time = self.recv('jail')
            outflow_rates = self.__compute_outflow_rates( message_time, 'jail' )
            self.send( (message_time, outflow_rates), 'jail' )

            # Interactions in the adjudication port
            #------------------------------------
            # one way "from" adjudication

            self.send( time, 'adjudication' )
            (check_time, inflow_rates) = self.recv('adjudication')
            assert abs(check_time-time) <= 1e-6
            self.ode_params['adjudication-inflow-rates'] = inflow_rates

            # Interactions in the arrested port
            #----------------------------------
            # one way "from" arrested

            self.send( time, 'arrested' )
            (check_time, inflow_rates) = self.recv('arrested')
            assert abs(check_time-time) <= 1e-6
            self.ode_params['arrested-inflow-rates'] = inflow_rates

            # Interactions in the community port
            #------------------------------
            # one way "to" community

            message_time = self.recv('community')
            outflow_rates = self.__compute_outflow_rates( message_time, 'community' )
            self.send( (message_time, outflow_rates), 'community' )

            # Evolve probation group population to the next time stamp
            #---------------------------------------------------------

            time = self.__step( time )

    def __rhs_fn(self, u_vec, t, params):

        fbg = u_vec  # probation population groups

        arrested_inflow_rates     = params['arrested-inflow-rates']
        adjudication_inflow_rates = params['adjudication-inflow-rates']

        inflow_rates  = arrested_inflow_rates + adjudication_inflow_rates

        cb0g = self.ode_params['commit-to-community-coeff-grps']

        cbjg = self.ode_params['commit-to-jail-coeff-grps']

        outflow_rates = ( cb0g + cbjg ) * fbg

        death_rates_coeff = params['probation-death-rates-coeff']

        death_rates = death_rates_coeff * fbg

        dt_fbg = inflow_rates - outflow_rates - death_rates

        return dt_fbg

    def __step(self, time=0.0):
        r'''
        ODE IVP problem:
        Given the initial data at :math:`t=0`,
        :math:`u = (u_1(0),u_2(0),\ldots)`
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.

        Parameters
        ----------
        time: float
            Time in the droplet unit of time (seconds).

        Returns
        -------
        None

        '''

        u_vec_0 = self.population_phase.GetValue('fbg', time)
        t_interval_sec = np.linspace(0.0, self.time_step, num=2)

        (u_vec_hist, info_dict) = odeint(self.__rhs_fn,
                                         u_vec_0, t_interval_sec,
                                         args=( self.ode_params, ),
                                         rtol=1e-4, atol=1e-8, mxstep=200,
                                         full_output=True)

        assert info_dict['message'] =='Integration successful.', info_dict['message']

        u_vec = u_vec_hist[1,:]  # solution vector at final time step
        values = self.population_phase.GetRow(time) # values at previous time

        time += self.time_step

        self.population_phase.AddRow(time, values)

        # Update current values
        self.population_phase.SetValue('fbg', u_vec, time)

        return time

    def __compute_outflow_rates(self, time, name):

        fbg = self.population_phase.GetValue('fbg',time)

        assert np.all(fbg>=0.0), 'values: %r'%fbg

        if name == 'jail':

            cbjg = self.ode_params['commit-to-jail-coeff-grps']

            outflow_rates = cbjg * fbg

        if name == 'community':

            cb0g = self.ode_params['commit-to-community-coeff-grps']

            outflow_rates = cb0g * fbg

        return outflow_rates

    def __zero_ode_parameters(self):
        '''
        If ports are not connected the corresponding outflows must be zero.

        '''

        zeros = np.zeros(self.n_groups)

        p_names = [p.name for p in self.ports]

        if 'community' not in p_names:
            self.ode_params['commit-to-community-coeff-grps']     = zeros

        if 'jail' not in p_names:
            self.ode_params['commit-to-jail-coeff-grps'] = zeros

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

import os
import pickle
import logging
from multiprocessing import Process

from cortix.src.module import Module
from cortix.src.port import Port

class Network:
    '''Cortix network.

    Attributes
    ----------

    num_networks: int
        Number of instances of this class.

    '''

    num_networks = 0

    def __init__(self):
        '''Module super class constructor.

        Attributes
        ----------
          max_n_modules_for_data_copy_on_root: int
              When using MPI the `network` will copy the data from all modules on the
              root process. This can generate an `out of memory` condition. This variable
              sets the maximum number of processes for which the data will be copied.
              Default is 1000.

       '''

        self.id = Network.num_networks

        self.name = 'network-'+str(self.id)
        self.log = logging.getLogger('cortix')

        self.max_n_modules_for_data_copy_on_root = 1000

        self.modules = list()

        self.gv_edges = list()

        self.use_mpi = None
        self.use_multiprocessing = None

        self.rank = None
        self.size = None
        self.comm = None

        Network.num_networks += 1

        return

    def module(self, m):
        '''Add a module.

        '''

        assert isinstance(m, Module), 'm must be a module'

        name = m.name
        if not name:
            name = m.__class__.__name__

        if m not in self.modules:
            m.use_mpi = self.use_mpi
            m.use_multiprocessing = self.use_multiprocessing
            self.modules.append(m)
            m.id = len(self.modules)-1  # see module doc for module id

        return

    def add_module(self, m):
        '''Alternative name to `module()`.

        '''

        self.module(m)

        return

    def connect(self, module_port_a, module_port_b, info=None):
        '''Connect two modules using either their ports directly or inferred ports.

        A connection always opens a channel for data communication in both ways.
        That is, both sends and receives are allowed.

        Note
        ----
        The simplest form of usage is with arguments: (`module_a`, `module_b`).
        In this case, a `port` with the name of `module_a` **must** exist in `module_b`,
        and vice-versa (port names as `str` in lower case). In addition, the connect
        must not be called again with these same two modules, else the underlying
        connection will be overriden.

        For more rigorous connection, the user is advised to fully specify the
        module and the port in each list argument.

        Parameters
        ----------
        module_port_a: list([Module,Port]) or list([Module,str]) or Module
            First `module`-`port` to connect.

        module_port_b: list([Module,Port]) or list([Module,str]) or Module
            Second `module`-`port` to connect.

        info: str
            Information on the directionality of the information flow. This is for
            graph visualization purposes only. The default value will use the order
            in the argument list to define the direction. Default: None. If set
            to `bidiretional`, will create a double headed arrow in the graph figure.

        '''

        if info:
            assert isinstance(info, str)

        if isinstance(module_port_a, Module) and isinstance(module_port_b, Module):

            module_a = module_port_a
            module_b = module_port_b

            assert module_a.name and module_b.name  # sanity check

            assert module_a in self.modules, 'module %r not in network.'%module_a.name
            assert module_b in self.modules, 'module %r not in network.'%module_b.name

            idx_a = self.modules.index(module_a)
            idx_b = self.modules.index(module_b)
            self.gv_edges.append((str(idx_a), str(idx_b)))

            if info == 'bidirectional':
                self.gv_edges.append((str(idx_b), str(idx_a)))

            port_a = module_a.get_port(module_b.name.lower())
            port_b = module_b.get_port(module_a.name.lower())

            port_a.connect(port_b)

        elif isinstance(module_port_a, list) and isinstance(module_port_b, list):

            assert len(module_port_a) == 2 and isinstance(module_port_a[0], Module) and\
                  (isinstance(module_port_a[1], str) or isinstance(module_port_a[1], Port))

            assert len(module_port_b) == 2 and isinstance(module_port_b[0], Module) and\
                  (isinstance(module_port_b[1], str) or isinstance(module_port_b[1], Port))

            module_a = module_port_a[0]
            module_b = module_port_b[0]

            assert module_a.name and module_b.name  # sanity check

            assert module_a in self.modules, 'module %r not in network.'%module_a.name
            assert module_b in self.modules, 'module %r not in network.'%module_b.name

            idx_a = self.modules.index(module_a)
            idx_b = self.modules.index(module_b)
            self.gv_edges.append((str(idx_a), str(idx_b)))

            if info == 'bidirectional':
                self.gv_edges.append((str(idx_b), str(idx_a)))

            if isinstance(module_port_a[1], str):
                port_a = module_a.get_port(module_port_a[1])
            elif isinstance(module_port_a[1], Port):
                port_a = module_port_a[1]
            else:
                assert False, 'help!'

            if isinstance(module_port_b[1], str):
                port_b = module_b.get_port(module_port_b[1])
            elif isinstance(module_port_b[1], Port):
                port_b = module_port_b[1]
            else:
                assert False, 'help!'

            port_a.connect(port_b)

        else:
            assert False, ' not implemented.'

        return

    def __run(self):
        '''
        Internal method to run the network simulation. Do not use this method, it is
        intended for Cortix to run it.

        This function concurrently executes the `cortix.src.module.run` function
        for each module in the network. Modules are run using either MPI or
        Multiprocessing, depending on the user configuration.

        Note
        ----
        When using multiprocessing, data from the modules state are copied to the master
        process after the `__run()` method of the modules is finished.

        '''
        assert len(self.modules) >= 1, 'the network must have a list of modules.'

        # Remove the scratch file save directory
        if self.rank == 0 or self.use_multiprocessing:
            os.system('rm -rf .ctx-saved && mkdir .ctx-saved')

        # Running under MPI
        #------------------
        if self.use_mpi:

            # Synchronize in the beginning
            assert self.size == len(self.modules) + 1,\
                'Incorrect number of processes (Required %r, got %r)'%\
                (len(self.modules) + 1, self.size)
            self.comm.Barrier()

            # Assign an mpi rank to all ports of a module using the module list index
            # If a port has rank assignment from a previous run; leave it alone
            for mod in self.modules:
                rank = self.modules.index(mod)+1
                for port in mod.ports:
                    if port.rank is None:
                        port.rank = rank

            # Assign a unique port id to all ports
            # If a port has id assignment from a previous run; leave it alone
            i = 0
            for mod in self.modules:
                for port in mod.ports:
                    port.use_mpi = self.use_mpi
                    if port.id is None:
                        port.id = i
                        i += 1

            # Parallel run module in MPI
            if self.rank != 0:
                mod = self.modules[self.rank-1]
                self.log.info('Launching Module {}'.format(mod))
                mod.run_and_save()

            # Sync here at the end
            self.comm.Barrier()

        # Running under Python multiprocessing
        #-------------------------------------
        else:

            # Parallel run all modules in Python multiprocessing
            processes = list()

            for mod in self.modules:
                self.log.info('Launching Module {}'.format(mod))
                proc = Process(target=mod.run_and_save)
                processes.append(proc)
                proc.start()

            # Synchronize at the end
            for proc in processes:
                proc.join()

        # Reload saved modules
        #---------------------
        if self.use_mpi:
            # make double sure all are in sync here before reading files from disk
            self.comm.Barrier()

        num_files = 0
        for file_name in os.listdir('.ctx-saved'):
            if file_name.endswith('.pkl'):
                num_files += 1
                file_name = os.path.join('.ctx-saved', file_name)
                with open(file_name, 'rb') as fin:
                    module = pickle.load(fin)
                    # reintroduce logging
                    module.log = logging.getLogger('cortix')
                    self.modules[module.id] = module

        if num_files and num_files != len(self.modules):
            self.log.warning('Network::run(): not all modules reloaded from disk.')

        if self.use_mpi:
            # Make double sure all are in sync here before going forward
            # this solves the problem of processes running behind reading files
            # that do not exist anymore
            self.comm.Barrier()

    def draw(self, graph_attr=None, node_attr=None, engine='twopi', lr=False,
             ports=False, node_shape='hexagon'):

        # Import here to avoid broken dependency. Only this method needs this.
        from graphviz import Digraph

        if graph_attr is None:
            graph_attr = {'splines':'true', 'overlap':'scale', 'ranksep':'2.0'}

        if node_attr is None:
            node_attr = {'shape':'hexagon', 'style':'filled'}

        dgr = Digraph(name=self.name, comment='Network::graphviz', format='png',
                      graph_attr=graph_attr, node_attr=node_attr, engine=engine)

        if lr:
            dgr.attr(rankdir='LR')

        dgr.attr('node', shape=node_shape)

        for idx, mod in enumerate(self.modules):
            dgr.node(str(idx), mod.name)

        for edg in self.gv_edges:
            dgr.edge(edg[0], edg[1])

        dgr.render()

        return dgr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
This is a simple way to hide the name of species of interest in a simulation.
The user would modify and copy this class into the Cortix module of interest
and keep it private.
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda
Sat Aug 15 13:41:12 EDT 2015
'''
#*********************************************************************************
import os
import sys
#*********************************************************************************

class Actor():
    '''
    See atoms list in Specie.
    '''

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__( self, name ):

        assert isinstance(name, str)

        # ** this is the section to be modified by the user of this class **
        self.__name_atoms_formula = {
              'water16'   : [['2*H-1','O-16'],'H2O'],
              'spc(v)'    : [['2*O-16'],'O2'],
              'spc1(v)'   : [['O-16'],'O'],
              'spc2(v)'   : [['Xe-136'],'Xe'],
              'spc3(v)'   : [['I-127'],'I'],
               }

        # ** do not modify beyond this point **

        assert name in self.__name_atoms_formula.keys(), 'name %r not valid.' % name

        self.__atoms = self.__name_atoms_formula[name][0]

        formula = self.__name_atoms_formula[name][1]
        assert isinstance(formula, str), 'formula %r not valid.' % formula
        self.__formula = formula

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def __get_atoms(self):

        '''
        Returns an ordered list of the number and type of all atoms that make
        up the actor. For example, ['2*H1', 'O16'].

        Returns
        -------
        atoms: list(str)
        '''

        return self.__atoms
    atoms = property(__get_atoms, None, None, None)

    def __get_formula(self):

        '''
        Returns the formula of the chemical in question. For example, 'H2O.

        Returns
        -------
        formula: str
        '''

        return self.__formula
    formula = property(__get_formula, None, None, None)

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

#======================= end class Actor =========================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org
'''
Heat exchanger used for reactor startup, shutdown and emergency operation.
'''

import logging

import scipy.constants as unit
import numpy as np

from cortix import Module

class Cooler(Module):
    """RCIS used in emergencies, cold startup and shutdown.
    Cools the reactor while the condenser is offline.
    """

    def __init__(self, params):
        """Constructor. No quantities needed at the moment.
        """

        super().__init__()

        self.port_names_expected = ['coolant-inflow', 'coolant-outflow', 'signal-in']

        self.params = params

        self.status = 'online'

        self.initial_time = params['start-time']
        self.end_time = params['end-time']
        self.time_step = 10.0

        self.show_time = (False, 10.0)

        self.log = logging.getLogger('cortix')
        self.__logit = False

        self.rcis_ua = 100000000 #w/m-k

    def run(self, *args):

        # Some logic for logging time stamps
        if self.initial_time + self.time_step > self.end_time:
            self.end_time = self.initial_time + self.time_step

        time = self.initial_time
        #unused print_time = self.initial_time
        print_time_step = self.show_time[1]
        if print_time_step < self.time_step:
            print_time_step = self.time_step

        while time <= self.end_time:

            #last_time_stamp = time - self.time_step

            #if self.show_time[0] and time >= print_time and \
            #        time<print_time+print_time_step:

            # Communicate information
            #------------------------
            self.__call_ports(time)

            # Evolve one time step
            #---------------------

            time = self.__step(time)

    def __call_ports(self, time):

        if self.status == 'online':

            # Interactions in the coolant-inflow port
            #-----------------------------------------
            # one way "from" coolant-inflow

            if self.get_port('coolant-inflow').connected_port:

                self.send(time, 'coolant-inflow')
                (check_time, inflow_state) = self.recv('coolant-inflow')
                assert abs(check_time-time) <= 1e-6
                inflow_temp = inflow_state['temp']
                flowrate = inflow_state['flowrate'] # kg/s


            # Interactions in the coolant-outflow port
            #-----------------------------------------
            # one way "to" coolant-outflow

            if self.get_port('coolant-outflow').connected_port:

                message_time = self.recv('coolant-outflow')

                outflow_temp = self.__get_coolant_outflow(message_time,
                                                          inflow_temp,
                                                          flowrate, self.params)

                self.send(outflow_temp, 'coolant-outflow')

        # Interactions in the signal-in port
        #-----------------------------------------
        # one way "to" signal-in

        if self.get_port('signal-in').connected_port:

            self.send(time, 'coolant-inflow')
            (check_time, rcis_state) = self.recv('signal-in')
            assert abs(check_time-time) <= 1e-6
            rcis_state = self.recv('signal-in')
            self.status = rcis_state
            print('signal is being run!', time)

    def __step(self, time=0.0):
        time += self.time_step
        return time

    def __get_coolant_outflow(self, time, temp_in, flowrate, params):
        """Calculate the temperature of the coolant leaving the RCIS system.

        Parameters
        ----------
        time: float
            Time in SI unit.
        temp_in: float
            The temperature of the coolant entering the RCIS system (also the current reactor
            temperature, in Kelvin).
        flowrate: float
            The current recirculation mass flowrate in kg/s.

        Returns
        -------
        temp_out: float
            The temperature of the cooled recirculation being fed back to the
            top of the reactor, in Kelvin.

        """
        heat_transfer_coef = self.rcis_ua
        cp_rho = 4.184 * unit.kilo * flowrate # Kj/Kg-k
        tc = 287.15 # assumed constant for simplicity

        if self.params['operating-mode'] == 'shutdown':
            heat_transfer_coef = heat_transfer_coef/10

        #initial guess: outflow temp = 300k
        temp_out_2 = temp_in - 5
        temp_out_1 = 999

        while abs(temp_out_2 - temp_out_1) > 0.1:
            temp_out_1 = temp_out_2
            delta_t1 = temp_in - tc
            delta_t2 = temp_out_1 - tc
            lmtd = (delta_t1 - delta_t2)/np.log(delta_t1/delta_t2)
            quantity = heat_transfer_coef * lmtd / cp_rho
            temp_out_2 = temp_in - quantity
            if temp_out_2 < tc:
                temp_out_2 = tc
                break

        return temp_out_2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

import os
import logging
import time
import datetime

from cortix.src.network import Network

class Cortix:
    '''Cortix main class definition.

    The typical Cortix run file workflow:

    1. Create the `Cortix` object
    2. Create tne (nested) network of modules
    3. Run and close `Cortix`

    '''

    def __init__(self, use_mpi=False, splash=False, log_filename='cortix'):
        '''Construct a Cortix simulation object.

        Parameters
        ----------
        use_mpi: bool
            True for MPI, False for multiprocessing.
        splash: bool
            Show the Cortix splash image.

        Attributes
        ----------
        network: Network
             A network of modules and their connectivity.
        use_mpi: bool
            `True` for MPI, `False` for Multiprocessing.
        use_multiprocessing: bool
            `False` for MPI, `True` for Multiprocessing.
        splash: bool
            Show the Cortix splash image.
        comm: mpi4py.MPI.Intracomm
            MPI.COMM_WORLD (if using MPI else None).
        rank: int
            The current MPI rank (if using MPI else None).
        size: int
            size of the group associated with MPI.COMM_WORLD.

        '''
        self.use_mpi = use_mpi
        self.use_multiprocessing = not use_mpi
        self.comm = None
        self.rank = None
        self.size = None

        self.splash = splash

        self.__network = None
        self.log_filename = log_filename
        # Fall back to multiprocessing if mpi4py is not available
        if self.use_mpi:
            try:
                from mpi4py import MPI
                self.comm = MPI.COMM_WORLD
                self.rank = self.comm.Get_rank()
                self.size = self.comm.size
            except ImportError:
                self.use_mpi = False

        # Setup the global logger 
        self.__create_logger()

        # Done
        if self.rank == 0 or self.use_multiprocessing:

            if self.splash:
                self.log.info('Created Cortix object %s', self.__get_splash(begin=True))
            else:
                self.log.info('Created Cortix object')

            # Initialize all date and timings
            self.wall_clock_time_start = time.time()
            self.wall_clock_time_end = self.wall_clock_time_start
            self.end_run_date = datetime.datetime.today().strftime('%d%b%y %H:%M:%S')
            os.system("rm -rf .ctx-saved && mkdir .ctx-saved")

        return

    def __set_network(self, n):
        assert isinstance(n,Network)
        n.use_mpi = self.use_mpi
        n.use_multiprocessing = self.use_multiprocessing
        n.rank = self.rank
        n.size = self.size
        n.comm = self.comm
        self.__network = n
        return

    def __get_network(self):
        return self.__network
    network = property(__get_network, __set_network, None, None)

    def run(self, save=False):
        '''Run the Cortix network simulation.

        '''

        self.__network._Network__run()

        if self.rank==0 or self.use_multiprocessing:
            self.wall_clock_time_end = time.time()
            self.log.info('run()::Elapsed wall clock time [s]: '+
                    str(round(self.wall_clock_time_end-self.wall_clock_time_start,2)))

    def close(self):
        '''Closes the cortix object properly before destruction.

        User is strongly advised to call this method at the end of the run file otherwise
        timings will not be recorded.

        '''

        # Sync here before close
        if self.use_mpi:
            self.comm.Barrier()

        if self.rank == 0 or self.use_multiprocessing:

            if self.splash:
                self.log.info('Closed Cortix object.'+self.__get_splash(end=True))
            else:
                self.log.info('Closed Cortix object.')

            self.wall_clock_time_end = time.time()

            self.log.info('close()::Elapsed wall clock time [s]: '+
                    str(round(self.wall_clock_time_end-self.wall_clock_time_start,2)))
        return

    def __create_logger(self):
        '''A helper function to setup the logging facility.'''

        # File removal
        if self.rank == 0 or self.use_multiprocessing:
            if os.path.isfile('cortix.log'):
                os.system('rm -rf cortix.log')

        # Sync here to allow for file removal
        if self.use_mpi:
            self.comm.Barrier()

        self.log = logging.getLogger(self.log_filename)

        self.log.setLevel(logging.DEBUG)
        #10/8/19 Added check to see if log hander exsists before creating new ones
        if not self.log.hasHandlers():
            file_handler = logging.FileHandler(self.log_filename+'.log')
            file_handler.setLevel(logging.DEBUG)

            console_handler = logging.StreamHandler()
            console_handler.setLevel(logging.DEBUG)

            # Formatter added to handlers
            if self.use_mpi:
                fs = '[rank:{}] %(asctime)s - %(name)s - %(levelname)s - %(message)s'.format(self.rank)
            else:

                fs = "[{}] %(asctime)s - %(name)s - %(levelname)s - %(message)s".format(os.getpid())

            formatter = logging.Formatter(fs)
            file_handler.setFormatter(formatter)
            console_handler.setFormatter(formatter)

            # Add handlers to logger
            self.log.addHandler(file_handler)
            self.log.addHandler(console_handler)

        return

    def __get_splash(self, begin=None, end=None):
        '''Returns the Cortix splash logo.

        Note
        ----
        Call this internal method with one argument only. Either `begin=True` or
        `end=True`.

        Parameters
        ----------
        begin: bool
            True for the beginning message, false for the ending.

        Returns
        -------
        splash: str
            The Cortix splash logo.

        '''

        assert begin==None or end==None
        if begin:
            end=False
        elif end:
            begin=False


        splash = \
        '_____________________________________________________________________________\n'+\
        '      ...                                        s       .     (TAAG Fraktur)\n'+\
        '   xH88"`~ .x8X                                 :8      @88>\n'+\
        ' :8888   .f"8888Hf        u.      .u    .      .88      %8P      uL   ..\n'+\
        ':8888>  X8L  ^""`   ...ue888b   .d88B :@8c    :888ooo    .     .@88b  @88R\n'+\
        'X8888  X888h        888R Y888r ="8888f8888r -*8888888  .@88u  ""Y888k/"*P\n'+\
        '88888  !88888.      888R I888>   4888>"88"    8888    ''888E`    Y888L\n'+\
        '88888   %88888      888R I888>   4888> "      8888      888E      8888\n'+\
        '88888 `> `8888>     888R I888>   4888>        8888      888E      `888N\n'+\
        '`8888L %  ?888   ! u8888cJ888   .d888L .+    .8888Lu=   888E   .u./"888&\n'+\
        ' `8888  `-*""   /   "*888*P"    ^"8888*"     ^%888*     888&  d888" Y888*"\n'+\
        '   "888.      :"      "Y"          "Y"         "Y"      R888" ` "Y   Y"\n'+\
        '     `""***~"`                                           ""\n'+\
        '                             https://cortix.org                              \n'+\
        '_____________________________________________________________________________'

        if begin:
            message = \
            '\n_____________________________________________________________________________\n'+\
            '                             L A U N C H I N G                               \n'

        else:
            message = \
            '\n_____________________________________________________________________________\n'+\
            '                           T E R M I N A T I N G                             \n'

        return message + splash

    def __del__(self):
        '''Destructs a Cortix simulation object.

        Warning
        -------
        By the time the body of this function is executed, the machinery of
        variables may have been deleted already. For example, `logging` is no longer
        there; do the least amount of work here.

        '''

        pass

if __name__ == '__main__':
    c = Cortix()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

import scipy.constants as const
from cortix.support.periodictable import ELEMENTS

class Species:
    '''
    All SI units (kg,s,K,Pa,J,W).

    The Specie() class encapsulates either the molecular or empirical chemical
    formula of a compound.
    This is done as follows. Say MAO2 is either a molecular or empirical chemical
    formula of a ficticious compound denoting minor actinides dioxide. The list
    of atoms is given as follows:

    ['0.49*Np-237', '0.42*Am-241', '0.08*Am-243', '0.01*Cm-244', '2.0*O-16']

    note the MA forming nuclides add to 1 = 0.49 + 0.42 + 0.08 + 0.01. Therefore
    the number of atoms in this compound is 3. 1 MA "ficticious" atom and 2 O.
    Note that the total number of "atoms" is obtained by summing all multipliers:
    0.49 + 0.42 + 0.08 + 0.01 + 2.0.
    The nuclide is indicated by the element symbol followed by a dash and the
    atomic mass number. Here the number of nuclide types is 5 (self.num_nuclide_types).

    The numbers preceeding the nuclide symbol before the * will be referred to as
    multipliers. The sum of the multipliers will add to the number of "atoms" in
    the formula. WARNING: a multiplier could be in the format 0.00e-00. In this
    case a hiphen may appear twice, e.g.: 1.549e-09*U-233

    Other forms can be used for common true species

    ['Np-237', '2.0*O-16'] or ['Np-237', 'O-16', 'O-16'] or [ '2*H', 'O' ] or
    [ 'H', 'O', 'H' ]  etc...

    This code will calculate the molar mass of any species with a given valid
    atom list using a provided periodic table of chemical elements. The user
    can also reset the value of the molar mass with a setter method.
    '''

    def __init__( self,
                  name='null-species-name',
                  formula_name='null-species-formula-name',
                  atoms=list(),
                  flag='null-species-flag',
                  info=None):

        assert isinstance(name, str)
        self.name = name

        assert isinstance(formula_name, str)
        self.formula_name = formula_name

        assert isinstance(atoms, list)
        self.atoms = atoms

        self.flag = flag  # flag can be any type

        self.info = info # info text such as technical name or other properties info

        self.molar_mass = 0.0      # kg/mol
        self.molar_heat_pwr = 0.0
        self.molar_gamma_pwr = 0.0
        self.molar_radioactivity = 0.0

        self.molar_mass_unit = 'kg/mol'

        self.molar_heat_pwr_unit = 'W/mol'
        self.molar_gamma_pwr_unit = 'W/mol'
        self.molar_radioactivity_unit = 'Ci/mol'

        self.molar_radioactivity_fractions = list()

        self.update_molar_mass()

        return

    def update_molar_mass(self):
        '''
        Updates the molar mass of the species after the molecular formula has
        been changed.

        '''

        molar_mass_const = const.physical_constants['molar mass constant'][0]
        molar_mass_const_unit = const.physical_constants['molar mass constant'][1]

        for entry in self.atoms:
            assert isinstance(entry, str)
            tmp = entry.split('*')
            nuclide = tmp[-1]
            element = nuclide.split('-')[0]
            assert element in ELEMENTS, 'element = %r'%(element)

        self.num_atoms = 0
        self.num_nuclide_types = 0

        nuclides = dict()

        num_atoms = 0
        summ = 0.0
        for entry in self.atoms:
            assert isinstance(entry, str)
            # format example:  3.2*O-18, or 3*O or O or O-16
            tmp = entry.split('*')
            multiple = 1.0
            # single nuclide
            if len(tmp) == 1:
                nuclide = tmp[0]
            # multiple nuclide
            elif len(tmp) == 2:
                multiple = float(tmp[0])
                nuclide = tmp[1]
            else:
                assert False

            nuclides[nuclide] = multiple
            num_atoms += multiple

            try:
                tmp = nuclide.split('-')
                if len(tmp) == 1:
                    element = ELEMENTS[tmp[0]]
                    rel_atomic_mass = element.exactmass  # from isotopic composition
                    if rel_atomic_mass == 0.0:
                        rel_atomic_mass = element.mass
                    molar_mass = rel_atomic_mass * molar_mass_const
                elif len(tmp) == 2:
                    element = ELEMENTS[tmp[0]].isotopes[int(tmp[1].strip('m'))]
                    molar_mass = element.mass * molar_mass_const
                else:
                    assert False
            except KeyError:
                summ += multiple * 0.0
            else:
                summ += multiple * molar_mass

        self.molar_mass = summ

        self.num_atoms = num_atoms
        self.num_nuclide_types = len(nuclides)

        return

    def reorder_formula(self):
        '''
        Takes a list of atoms for a molecular or empirical formula and places
        it in order of decreasing magnitude of stoichiometric coefficient. For
        example, [O, 2*H] will be returned as [2*H, O]. This is used for
        printing purposes. The internal order will not change.

        Returns
        -------
        atoms2: list

        '''

        atoms1 = self.atoms[:]  # shallow copy
        atoms2 = list()

        if len(self.atoms) <= 1:
            return atoms1

        if len(self.atoms) > 1:

            # save the multiplier value as a string type of scientific notation
            for entry in self.atoms:

                assert isinstance(entry, str)

                # format example:  3.2*O-18, or 3*O or O or O-16
                tmp = entry.split('*')

                nuclide = tmp[-1]

                multiplier = 0.0

                if len(tmp) == 1:
                    multiplier = 1
                elif len(tmp) == 2:
                    multiplier = float(tmp[0])
                else:
                    assert False

                assert multiplier != 0.0, 'multiplier = %r' % (multiplier)

                multiplier = '{0:9.3e}'.format(multiplier)

                atoms1[self.atoms.index(entry)] = multiplier + '*' + nuclide

            # order in decreasing order of multiplier magnitude
            multipliers_lst = list()

            for entry in atoms1:

                tmp = entry.split('*')

                multiplier = 0.0

                if len(tmp) == 1:
                    continue
                elif len(tmp) == 2:
                    multiplier = float(tmp[0])
                else:
                    assert False

                multipliers_lst.append(float(multiplier))

            sorted_atoms = [a for (i, a) in sorted(zip(multipliers_lst, atoms1),
                key=lambda pair: pair[0], reverse=True)]

            atoms2 = sorted_atoms

        return atoms2

    def __str__(self):
        s = '\n\t ' + \
            '\n\t Species(): name=%s;' + \
            ' formula_name=%s;' + \
            '\n\t formula=%s;' + \
            '\n\t # atoms=%s;' + ' # nuclide types=%s;' + ' molar mass=%9.3e[%s];' + \
            '\n\t flag=%s;' + \
            '\n\t info=%s;' + \
            '\n\t molar radioactivity=%9.3e[%s];' + \
            '\n\t molar heat pwr=%9.3e[%s];' + \
            '\n\t molar gamma pwr=%9.3e[%s];' + \
            '\n\t atoms=%s;' + \
            '\n\t molar radioactivity fractions=%s'
        return s % (self.name,
                self.formula_name,
                self.reorder_formula(),
                self.num_atoms, self.num_nuclide_types, self.molar_mass, \
                        self.molar_mass_unit,
                self.flag,
                self.info,
                self.molar_radioactivity, self.molar_radioactivity_unit,
                self.molar_heat_pwr, self.molar_heat_pwr_unit,
                self.molar_gamma_pwr, self.molar_gamma_pwr_unit,
                [i.split('*')[-1] for i in self.atoms],
                ['%9.3e' % i for i in self.molar_radioactivity_fractions])

    def __repr__(self):
        s = '\n\t Species(): name=%s;' + \
            ' formula_name=%s;' + \
            '\n\t formula=%s;' + \
            '\n\t # atoms=%s;' + ' # nuclide types=%s;' + ' molar mass=%9.3e[%s];' + \
            '\n\t flag=%s;' + \
            '\n\t molar radioactivity=%9.3e[%s];' + \
            '\n\t molar heat pwr=%9.3e[%s];' + \
            '\n\t molar gamma pwr=%9.3e[%s];' + \
            '\n\t atoms=%s;' + \
            '\n\t molar radioactivity fractions=%s'
        return s % (self.name,
                self.formula_name,
                self.reorder_formula(),
                self.num_atoms, self.num_nuclide_types, self.molar_mass, \
                        self.molar_mass_unit,
                self.flag,
                self.molar_radioactivity, self.molar_radioactivity_unit,
                self.molar_heat_pwr, self.molar_heat_pwr_unit,
                self.molar_gamma_pwr, self.molar_gamma_pwr_unit,
                [i.split('*')[-1] for i in self.atoms],
                ['%9.3e' % i for i in self.molar_radioactivity_fractions])

if __name__ == '__main__':
    tbp_org = Species( name='TBP', formula_name='(C4H9O)_3PO(o)',
              atoms=['12*C','27*H','4*O','P'] )
    print(tbp_org)
    no3Minus_aqu = Species( name='NO3-', formula_name='NO3-(a)',
                   atoms=['N','3*O'] )
    print(no3Minus_aqu)



import os, time, datetime, threading, random, sys, string, math
import numpy as np
from cortix import Module
from cortix import Network
from cortix import Cortix

class Messenger:
    def __init__(self, circle=None, collision = [], timestamp='0'):
        self.collision = []
        self.timestamp = timestamp
        self.m = 1
        self.r = 1
        self.v = [0.0,0.0]
        self.p = [0.0,0.0]
        self.name = ''
        self.ke = 0
        self.color = 'b'
        self.elapsed = 0
        self.total_collisions = 0

class Particle:
    def __init__(self,shape,bn=[0,0,20,20],color='b',r=1):
        super().__init__()
        coords = shape
        self.bndry = []
        self.t_step = 0.01
        for c,f in enumerate(coords):
            
            try:
                cr = (coords[c],coords[c+1])
            except IndexError:
                cr = (coords[c],coords[-1])
##            
            self.bndry.append(cr)
        self.r=r
        for i in range(100): #Attempt to spawn ball within boundary
            self.p0 = [random.uniform(bn[0],bn[2]),random.uniform(bn[1],bn[3])]
            break
        if i>85:
            print('Warning, ball took:',i,'attempts to spawn')
        self.v0 = [random.uniform(-50,50),random.uniform(-30,30)]
        self.cor = 0.60
        self.a = (0,0)
        self.m = 1
        self.ke = 0.5*self.m*((self.v0[0]**2+self.v0[1]**2)**0.5)**2
        self.timestamp=str(datetime.datetime.now())
        self.collision=[]
        self.name = ''.join([random.choice(string.ascii_lowercase+string.digits) for f in range(10)])
        #Customize container class that is sent to other modules
        self.messenger = Messenger()
        self.messenger.timestamp = self.timestamp
        self.messenger.m,self.messenger.r = self.m,self.r
        self.messenger.v = self.v0
        self.messenger.p = self.p0
        self.messenger.name = self.name
        self.messenger.color = color
        self.messenger.elapsed = 0
        self.messenger.ke = self.ke
        self.mycollisions = []

    def run(self, ball_list):
        self.ball_list = ball_list
        self.collisions=0
        t = self.t_step
        self.messenger.elapsed += t
        for ball in self.ball_list: #Detects collision with other objects
            if self.name==ball.name:
                continue
            #ball is Messenger class object
            dif = np.subtract(self.p0,ball.p)
            dis = math.sqrt(dif[0]**2 + dif[1]**2)
            
            name = ball.name
            if dis <= self.r + ball.r:# self.circle.crosses(shape) or self.circle.touches(shape) or self.circle.intersects(shape):
                coldic = dict(name=ball.name,v0=self.v0,p0=self.p0,elapsed=self.messenger.elapsed)
                self.mycollisions.append(coldic)
                self.messenger.collision.append(coldic)
                self.ball_collision(ball,ball.p,ball.v)
                self.ball_shift(ball,ball.p,ball.v)
                coldic['p0'] = self.p0
        
        for ball in self.ball_list: #Detects collision with other objects
            #Reacts to intersection between this object and another
            for c,line in enumerate(ball.collision): #Undetected Collisions received as a message
                if self.name == line['name'] and line not in self.mycollisions:
                    p0,v0 = line['p0'],line['v0']
                    dif = np.subtract(self.p0,p0)
                    dis = math.sqrt(dif[0]**2 + dif[1]**2)
##                    print('Collision%', self.name,dis)
                    self.ball_collision(ball,p0,v0)
                    if dis <= self.r + ball.r:
                        self.ball_shift(ball,p0,v0)
                    del ball.collision[c]
                for d, col in enumerate(self.mycollisions):
                    if line == col and col!= []:
                        del self.mycollisions[d]
                        del ball.collision[c]
        
                
        #Gravity calculations for timestep
        self.p0[1] = 0.5*self.a[1]*t**2+self.v0[1]*t+self.p0[1]
        self.p0[0] = 0.5*self.a[0]*t**2+self.v0[0]*t+self.p0[0]
        self.v0[1] = self.a[1]*t + self.v0[1]
        self.v0[0] = self.a[0]*t + self.v0[0]
        
        
        
                
        
        for (c1,c2) in self.bndry: #Detects collision with boundary
            dif = np.subtract(self.p0,c1)
            dis = math.sqrt(dif[0]**2 + dif[1]**2)
            clen = np.subtract(c1,c2)
            clen = math.sqrt(clen[0]**2 + clen[1]**2)
            angle = math.atan2(self.p0[1]-c1[1], self.p0[0]-c1[0]) - math.atan2(c2[1]-c1[1], c2[0]-c1[0])
            wall_distance = math.sin(angle)*dis
            closest_dis = math.cos(angle)*dis
            
            #Wall detection: https://stackoverflow.com/questions/1073336/circle-line-segment-collision-detection-algorithm
            if abs(wall_distance) <= self.r and 0<closest_dis <clen:
##                print('Wall Collision',(c1,c2))
    
                self.ke = 0.5*self.m*((self.v0[0]**2+self.v0[1]**2)**0.5)**2
##                print(self.name,'wall collision. Kinetic Energy: ',self.ke)
                self.wall_collision(c1,c2,wall_distance)
##                if self.ke != 0.5*self.m*((self.v0[0]**2+self.v0[1]**2)**0.5)**2:
##                    print("ENERGY CHANGED!!!",self.ke - 0.5*self.m*((self.v0[0]**2+self.v0[1]**2)**0.5)**2)
        self.messenger.v = self.v0
        self.messenger.p = self.p0
        self.ke = 0.5*self.m*((self.v0[0]**2+self.v0[1]**2)**0.5)**2
        self.messenger.ke = self.ke
        return self.messenger

    def wall_collision(self,c1,c2,d):
        angle3 = np.arctan2(c2[1] - c1[1], c2[0] - c1[0])+np.pi/2
        self.p0 = [self.p0[0]-(self.r+d)*np.cos(angle3), self.p0[1]-(self.r+d)*np.sin(angle3)]
        angle2 = np.arctan2(self.v0[1], self.v0[0])
        v = (self.v0[0]**2+self.v0[1]**2)**0.5
        theta = (angle2-angle3)
        vbi, vbj = v*np.sin(theta), v*np.cos(theta)
        vbj = -vbj *self.cor
        v = (vbi**2+vbj**2)**0.5
        angle4 = np.arctan2(vbj, vbi)
        angle1 = angle4-angle3
        self.collisions+=1
        self.v0 = [np.sin(angle1)*v, np.cos(angle1)*v]
        self.messenger.total_collisions += 1
        
    def ball_shift(self,ball,p,v):
        angle = np.arctan2(p[1] - self.p0[1], p[0] - self.p0[0])
        pnp = np.array(p)
        dif = np.subtract(self.p0,p)
        d = math.sqrt(dif[0]**2 + dif[1]**2)
        self.p0 = [self.p0[0]-(self.r+ball.r-d)*np.cos(angle),self.p0[1]-(self.r+ball.r-d)*np.sin(angle)]
        
    def ball_collision(self,messenger,p0,v0):
##        print('A ball collision',self.name,messenger.name)
        v2,m = v0,messenger.m
        v3 = (v2[0]**2+v2[1]**2)**0.5
        phi = np.arctan2(v2[1],v2[0])
        p2x,p2y = p0[0],p0[1]
        angle = np.arctan2(p2y - self.p0[1], p2x - self.p0[0]) 
        angle2 = np.arctan2(self.v0[1], self.v0[0])
        v = (self.v0[0]**2+self.v0[1]**2)**0.5
        
        #Equation source: https://en.wikipedia.org/wiki/Elastic_collision
        vpx=((v*np.cos(angle2-angle)*(self.m-m)+2*m*v3*np.cos(phi-angle))/(self.m+m))*np.cos(angle)+v*np.sin(angle2-angle)*np.cos(angle+np.pi/2)
        vpy=((v*np.cos(angle2-angle)*(self.m-m)+2*m*v3*np.cos(phi-angle))/(self.m+m))*np.sin(angle)+v*np.sin(angle2-angle)*np.sin(angle+np.pi/2)
        vp = (vpx**2+vpy**2)**0.5
        self.v0 = [vpx,vpy]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda

Fuel segment

VFdALib support classes

Sat Jun 27 14:46:49 EDT 2015
'''
#*********************************************************************************
import os
import sys
import io
import time
import datetime
import math
import random

from cortix.support.nuclear.fuel_segment import FuelSegment
#*********************************************************************************

class FuelSegmentsGroups():
    '''
    Creates a dictionary of lists of fuel segment objects, with the keys
    typically being timestamps. Each fuel segment object has two data members,
    a `Pandas` Series for geometry spec and a panda DataFrame for property
    density.
    '''

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self,
                 key=None,
                 fuelSegments=None
                 ):

        # constructor
        # FuelSegmentsGroups simply encapsulates a dictionary of a list of
        # FuelSegment objects. The key is typically a time stamp.
        # A FuelSegment object has two data members, a pandas Series for geometry spec
        # and a pandas DataFrame for property density.

        # This is the central member data
        self.groups = dict()

        assert key not in self.groups.keys()

        if fuelSegments is not None:
            assert isinstance(fuelSegments, list)
            assert isinstance(fuelSegments[-1], FuelSegment)
            self.groups[key] = fuelSegments
        else:
            self.groups[key] = list()

        return

#*********************************************************************************
# Public Member Functions
#*********************************************************************************

    def HasGroup(self, key):

        '''
        Checks if the specified key has a group of fuel segments associated
        with it.

        Parameters
        ----------
        key: str

        Returns
        -------
        key: str
        '''

        return key in self.groups.keys()

    def AddGroup(self, key, fuelSegments=None):

        '''
        Appends the dictionary with a new key and associated list of
        fuelSegments. If the specified key is already present in the
        dictionary, then the specified list of fuel segments will be appended
        to the list of fuel segments already associated with the specified key.

        Parameters
        ----------
        key: str
        fuelSegments: list
        '''

        self.__AddGroup(key, fuelSegments)

        return

    def GetAttribute(self, groupKey=None, attributeName=None,
                     nuclideSymbol=None, nuclideSeries=None):

        '''
        Returns the average value of an attribute amongst all elements in a
        group (WARNING: keys with no values associated with them will lower
        this average!). If groupKey is not specified, the function will return
        the average attribute value of every fuel segment element in the
        entire dictionary. If attribute is not specified, the function call
        will fail. If the key value specified does not match any keys in the
        dictionary, the function will return a value of 0.

        Parameters
        ----------
        groupKey: str
        attributeName: str
        nuclideSymbol: str
        nuclideSeries: str

        Returns
        -------
        groupAttribute: float
        '''

        return self.__GetGroupAttribute(
            groupKey, attributeName, nuclideSymbol, nuclideSeries)

    def GetFuelSegments(self, groupKey=None):

        '''
        Returns a list of fuel segments associated with a specified groupkey.
        If no group key is specified, then all elements in the dictionary
        will be returned. If the specified group key does not exist, then the
        function will return an empty list.

        Parameters
        ----------
        groupKey: str

        Returns
        ----------
        fuelSegments: list
        '''

        return self.__GetFuelSegments(groupKey)

    def RemoveFuelSegment(self, groupKey, fuelSegment):

        '''
        Removes a fuel segment from a list associated with a specified group
        key. If the specified group key or fuel segment do not exist, the
        function will fail.

        Parameters
        ----------
        groupKey: str
        fuelSegment: str

        Returns
        -------
        empty:
        '''

        return self.__RemoveFuelSegment(groupKey, fuelSegment)

# def __str__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formula, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)
#
# def __repr__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formula, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

# If an attibute is not found, Return 0 even if a groupKey is not found
# Don't change this behavior; it will break user's code.

    def __GetGroupAttribute(self, groupKey=None,
                            attributeName=None, symbol=None, series=None):

        '''
        Returns the cumulative or average densities of all fuel segments in all
        groups. Use this function with discretion, as groups with no segments
        will reduce the average density value.

        Parameters
        ----------
        groupKey: str
        attributeName: str
        symbol: str
        series: list

        Returns
        -------
        attribute: float
        '''

        assert attributeName is not None, 'fatal.'

        attribute = None

        if groupKey is None:

            attribute = 0

            for (key, fuelSegments) in self.groups.items():
                assert isinstance(fuelSegments, list), 'fail.'
                if len(fuelSegments) == 0:
                    continue  # this will reduce the average value
                groupAttribute = 0
                for fuelSegment in fuelSegments:
                    groupAttribute += fuelSegment.get_attribute(
                        attributeName, symbol, series)

                if attributeName[-4:] == 'Dens' or attributeName[-2:] == 'CC':
                    groupAttribute /= len(fuelSegments)
                attribute += groupAttribute

            if attribute != 0 and \
               (attributeName[-4:] == 'Dens' or attributeName[-2:] == 'CC'):
                attribute /= len(self.groups)

#     if attributeName[-4:] == 'Dens' or attributeName[-2:] == 'CC':
#        print('HELLO density '+attributeName+' ', attribute)

        # Get average property in all fuel segments within a groupKey
        else:

            if groupKey not in self.groups.keys():
                return 0

            fuelSegments = self.groups[groupKey]

            if len(fuelSegments) is 0:
                return 0

            attribute = 0

            for fuelSegment in fuelSegments:
                attribute += fuelSegment.get_attribute(
                    attributeName, symbol, series)

            if attribute != 0 and \
               (attributeName[-4:] == 'Dens' or attributeName[-2:] == 'CC'):
                attribute /= len(fuelSegments)

        return attribute

    def __GetFuelSegments(self, groupKey=None):

        '''
        Returns a list of fuel segments associated with a given group
        (if groupKey is specified), or an ordered list of pairs of all
        segments in all groups and their keys. [ (timeStamp, fuelSegment),
        (timeStamp, fuelSegment), ..]. If the specified groupKey does not
        exist, this function will return an empty list.

        Parameters
        ----------
        groupKey: str

        Returns
        -------
        sorted_data: type or self.groups[groupKey]: type or list(): list
        '''

        if groupKey is None:  # return an ordered list of all fuelSegments

            tmp = list()
            timeStamp = list()
            for (key, fuelSegments) in self.groups.items():
                if fuelSegments is None:
                    continue
                tmp += fuelSegments
                # all fuel segments in the group have
                timeStamp += [key for i in fuelSegments]
                # the same time stamp

            # sort fuelSegments in order of their keys
            data = zip(timeStamp, tmp)  # this is a list of pairs
            sorted_data = sorted(
                data, key=lambda entry: entry[0], reverse=False)
            #tmp = [ y for (x,y) in sorted_data ]  # oldest first
            return sorted_data

        else:

            if groupKey not in self.groups.keys():
                return list()

            return self.groups[groupKey]

    def __AddGroup(self, groupKey, fuelSegments=None):

        '''
        If a list of fuel segments and a group key is specified, the
        fuelSegment list is appended to the specified groupKey. If the groupKey
        specified does not already exist, a new one is created and the
        fuelSegments list is appended to it. fuelSegments is ALWAYS a list,
        and may be empty. A group will ALWAYS have a  fuelSegments list.

        Parameters
        ----------
        groupKey: str
        fuelSegments: list
        '''

        if fuelSegments is None:
            fuelSegments = list()
        else:
            assert isinstance(fuelSegments, list), 'fail.'

        if groupKey in self.groups.keys():
            self.groups[groupKey] += fuelSegments
        else:
            self.groups[groupKey] = fuelSegments

        return

    def __RemoveFuelSegment(self, groupKey, fuelSegment_remove):

        '''
        Removes a fuel segment from a list associated with a specified group
        key. If the specified group key or fuel segment do not exist, the
        function will fail.

        Parameters
        ----------
        groupKey: str
        fuelSegment: str
        '''

        assert groupKey in self.groups.keys(), 'fail.'

        fuelSegments = self.groups[groupKey]
        nSegments = len(fuelSegments)

        for fuelSegment in fuelSegments:
            if fuelSegment.get_attribute(
                    'segmentId') == fuelSegment_remove.get_attribute('segmentId'):
                fuelSegments.remove(fuelSegment)

        assert len(self.groups[groupKey]) == nSegments - 1, 'fatal.'

        return

#========================= end class FuelSegmentsGroups ==========================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org

import logging

import math
from scipy.integrate import odeint
import scipy.constants as unit
import numpy as np

import iapws.iapws97 as steam
import iapws.iapws95 as steam2

from cortix import Module
from cortix.support.phase_new import PhaseNew as Phase
from cortix import Quantity

class BWR(Module):
    """Boiling water reactor single-point reactor.

    Notes
    -----
    These are the `port` names available in this module to connect to respective
    modules: `turbine`, and `pump`.
    See instance attribute `port_names_expected`.

    """

    def __init__(self, params):
        """Constructor.

        Parameters
        ----------
        params: dict
            All parameters for the module in the form of a dictionary.

        """

        super().__init__()

        self.port_names_expected = ['coolant-inflow', 'coolant-outflow',
                                    'RCIS-inflow', 'RCIS-outflow']

        self.params = params

        self.initial_time = params['start-time']
        self.end_time = params['end-time']
        self.time_step = 10.0

        self.show_time = (False, 10.0)

        self.log = logging.getLogger('cortix')
        self.__logit = False

        #Data pertaining to one-group energy neutron balance
        self.gen_time = 1.0e-4  # s
        self.beta = 6.5e-3  # params['k_infty'] = 1.3447
        # geometric buckling; B = (pi/R)^2 + (2.405/H)^2
        self.buckling = (math.pi/237.5)**2.0 + (2.405/410)**2.0
        self.q_0 = 5000
        # fuel macroscopic absorption cross section, cm^-1
        self.fuel_macro_a = 1.34226126162
        # moderator microscopic absorption cross section, cm^2
        self.mod_micro_a = 0.332 * unit.zepto * unit.milli
        #number density of the fuel, atoms/cm^3
        self.n_fuel = 1.9577906e+21
        #resonance integral, I (dimensionless)
        self.I = 40.9870483 * unit.zepto * unit.milli
        # moderator microscopic scattering cross section, cm^2
        self.mod_micro_s = 20 * unit.zepto * unit.milli
        self.xi = 1 # average logarithmic energy decrement for light water
        # energy of a neutron produced by fissioning, in electron volts
        self.E0 = 2 * unit.mega
        self.mod_mu0 = 0.71 # migration and diffusion area constants
        self.eta = 1.03 # fast fission factor
        self.epsilon = 2.05 # neutron multiplecation factor
        self.mod_molar_mass = 18 # g/mol
        
        # Delayed neutron emission
        self.species_decay = [0.0124, 0.0305, 0.111, 0.301, 1.14, 3.01] # 1/sec
        self.species_rel_yield = [0.033, 0.219, 0.196, 0.395, 0.115, 0.042]

        # Data pertaining to two-temperature heat balances
        self.fis_energy = 180 * 1.602e-13 # J/fission
        self.enrich = 4.3/100.
        self.fuel_mat_mass_dens = 10.5 # g/cc
        #params['moderator_fuel_ratio'] = 387 # atomic number concentration ratio
        self.sigma_f_o = 586.2 * 100 * 1e-30 # m2
        self.temp_o = 20 + 273.15 # K
        self.thermal_neutron_velo = 2200 # m/s

        self.fis_nuclide_num_dens = 9.84e26 # (fissile nuclei)/m3

        self.q_c = 303 # volumetric flow rate

        self.fuel_dens = 10500 # kg/m3
        self.cp_fuel = 175 # J/(kg K)
        self.fuel_volume = 7.25  # m3

        self.steam_flowrate = 1820 # kg/s
        self.coolant_dens = 1000 #  kg/m3
        self.cp_coolant = 3500# J/(mol K) - > J/(kg K)
        self.coolant_volume = 7 #m3

        self.ht_coeff = 45000000

        # J/(fission sec) 1.26 t^-1.2 (t in seconds)
        self.fis_beta_energy = 1.26 * 1.602e-13
        # J/(fission sec) 1.40 t^-1.2 (t in seconds)
        self.fis_alpha_energy = 1.40 * 1.602e-13
        # % subcooling based on the % subcooling that exists at steady state
        self.subcooling_percent = 1 #(1 -(steam_table._Region4(7, 0)["h"]  - steam_table._Region1(493.15, 7)["h"])/(steam_table._Region4(7,0)["h"]))
        self.shutdown_temp_reached = False
        self.q_source_status = 'in' # is q_source inserted (in) or withdrawn (out)

        gen_time = self.gen_time # retrieve neutron generation time
        self.q_0 = 0.1

        self.n_ss = 0 # neutronless steady state before start up

        rho_0_over_beta = 0.25 # $
        beta = self.beta

        # control rod reactivity worth; enough to cancel out the negative
        self.alpha_n = 0

        self.reactivity = rho_0_over_beta * beta # "rho/beta = 10 cents"

        self.temp_o = self.temp_o

        self.tau = 1 # s
        self.malfunction_subcooling = 0.75
        self.alpha_n_malfunction = 0
        n_species = len(self.species_decay)

        assert len(self.species_rel_yield) == n_species

        # initialize concentration vector
        c_vec_0 = np.zeros(n_species, dtype=np.float64)
        # retrieve list of decay constants
        species_decay = self.species_decay
        lambda_vec = np.array(species_decay) # create a numpy vector
        beta = self.beta
        species_rel_yield = self.species_rel_yield
        # create the beta_i's vector
        beta_vec = np.array(species_rel_yield) * beta
        gen_time = self.gen_time # retrieve neutron generation time
        n_ss = self.n_ss
        # compute the steady state precursors number density
        c_vec_ss = beta_vec/lambda_vec/gen_time * n_ss
        self.c_vec_ss = c_vec_ss
        # setup initial condition for variables
        self.n_0 = n_ss
        self.c_vec_0 = c_vec_ss
        self.rho_0 = self.reactivity

        self.RCIS = False

        # Coolant outflow phase history
        quantities = list()

        flowrate = Quantity(name='flowrate', formal_name='q_c', unit='kg/s', value=0.0,
                            info='Reactor Outflow Coolant Flowrate', latex_name=r'$q_c$')

        quantities.append(flowrate)

        temp = Quantity(name='temp', formal_name='T_c', unit='K', value=params['coolant-temp'],
                        info='Reactor Outflow Coolant Temperature', latex_name=r'$T_c$')

        quantities.append(temp)

        press = Quantity(name='pressure', formal_name='P_c', unit='Pa', value=0.0,
                         info='Reactor Outflow Coolant Pressure', latex_name=r'$P_c$')

        quantities.append(press)

        quality = Quantity(name='steam-quality', formal_name='chi_s', unit='', value=0.0,
                           info='Reactor Steam Quality', latex_name=r'$\chi$')
        quantities.append(quality)

        self.coolant_outflow_phase = Phase(time_stamp=self.initial_time, time_unit='s', quantities=quantities)

        # Neutron phase history
        quantities = list()

        neutron_dens = Quantity(name='neutron-dens', formal_name='n', unit='1/m^3',
                                value=params['n-dens'], info='Rel. Reactor Neutron Density', latex_name=r'$n$')

        quantities.append(neutron_dens)

        delayed_neutron_cc = Quantity(name='delayed-neutrons-cc', formal_name='c_i',
                                      unit='1/m^3 ', value=params['delayed-neutron-cc'],
                                      info='Rel. Delayed Neutron Precursors', latex_name=r'$c_i$')
        quantities.append(delayed_neutron_cc)

        self.neutron_phase = Phase(time_stamp=self.initial_time, time_unit='s',
                                   quantities=quantities)

        # Reactor phase
        quantities = list()

        fuel_temp = Quantity(name='fuel-temp', formalName='T_f', unit='K',
                             value=params['fuel-temp'], info='Reactor Nuclear Fuel Temperature',
                             latex_name=r'$T_f$')

        quantities.append(fuel_temp)

        reg_rod_position = Quantity(name='reg-rod-position',
                                    formal_name='reg rod position', unit='m', value=0.0,
                                    info='Reactor Regulating Rod Position', latex_name=r'$x_p$')

        quantities.append(reg_rod_position)

        self.reactor_phase = Phase(time_stamp=self.initial_time, time_unit='s',
                                   quantities=quantities)

        # Initialize inflow
        self.params['inflow-cool-temp'] = params['condenser-runoff-temp']

        self.params['decay-heat-0'] = 0
        # Set up initial decay heat value
        if self.params['shutdown-mode']:
            temp = params['fuel-temp']
            n_dens = params['n-dens']
            decay_heat = self.__nuclear_pwr_dens_func(999, temp, n_dens, params)
            decay_heat *= 0.0622 # assume decay heat at 6.5% total power
            self.params['decay-heat-0'] = decay_heat

    def run(self, *args):
        # Some logic for logging time stamps
        if self.initial_time + self.time_step > self.end_time:
            self.end_time = self.initial_time + self.time_step

        time = self.initial_time
        print_time = self.initial_time
        print_time_step = self.show_time[1]
        if print_time_step < self.time_step:
            print_time_step = self.time_step

        while time <= self.end_time:

            #last_time_stamp = time - self.time_step

            #if self.show_time[0] and time >= print_time and \
            #        time<print_time+print_time_step:
            if self.show_time[0] and print_time <= time < print_time+print_time_step:

                msg = self.name+'::run():time[m]='+ str(round(time/unit.minute, 1))
                self.log.info(msg)

                self.__logit = True
                print_time += self.show_time[1]

            else:
                self.__logit = False

            # Communicate information
            #------------------------
            self.__call_ports(time)

            # Evolve one time step
            #---------------------

            time = self.__step(time)

    def __call_ports(self, time):

        current_temp = self.__get_coolant_outflow(time)['temp']
        if current_temp < 373.15:
            if self.params['operating-mode'] == 'shutdown':
                self.RCIS = True

        if current_temp > 373.15:
            if self.params['operating-mode'] == 'startup':
                self.RCIS = False

        # Interactions in the coolant-outflow port
        #-----------------------------------------
        # one way "to" coolant-outflow

        # send to
        if self.get_port('coolant-outflow').connected_port:
            message_time = self.recv('coolant-outflow')

            if self.RCIS:
                coolant_outflow = dict()
                coolant_outflow['temp'] = 287.15
                coolant_outflow['pressure'] = 0
                coolant_outflow['quality'] = 0
                coolant_outflow['flowrate'] = 0
                self.send((message_time, coolant_outflow), 'coolant-outflow')
            else:
                coolant_outflow = self.__get_coolant_outflow(message_time)
                #outflow_params = dict()
                self.send((message_time, coolant_outflow), 'coolant-outflow')

        # Interactions in the RCIS-outflow port
        #-----------------------------------------
        # one way "to" RCIS-outflow

        # send to
        if self.get_port('RCIS-outflow').connected_port:
            message_time = self.recv('RCIS-outflow')

            if self.RCIS:
                coolant_outflow = self.__get_coolant_outflow(message_time)
                coolant_outflow['flowrate'] = 5280 # kg/s
                self.send((message_time, coolant_outflow), 'RCIS-outflow')
            else:
                coolant_outflow = self.__get_coolant_outflow(message_time)
                coolant_outflow['flowrate'] = 0.0 # kg/s
                self.send((message_time, coolant_outflow), 'RCIS-outflow')

        # Interactions in the coolant-inflow port
        #----------------------------------------
        # one way "from" coolant-inflow

        # receive from
        if self.get_port('coolant-inflow').connected_port:
            self.send(time, 'coolant-inflow')

            if self.RCIS:
                (check_time, _) = self.recv('coolant-inflow')
                assert abs(check_time-time) <= 1e-6
            else:
                (check_time, inflow_cool_temp) = self.recv('coolant-inflow')
                assert abs(check_time-time) <= 1e-6
                self.params['inflow-cool-temp'] = inflow_cool_temp

        # Interactions in the RCIS-inflow port
        #----------------------------------------
        # one way "from" RCIS-inflow

        # receive from
        if self.get_port('RCIS-inflow').connected_port:
            self.send(time, 'RCIS-inflow')

            if self.RCIS:
                (check_time, inflow_temp) = self.recv('RCIS-inflow')
                assert abs(check_time-time) <= 1e-6
                self.params['inflow-cool-temp'] = inflow_temp
            else:
                (check_time, _) = self.recv('RCIS-inflow')
                assert abs(check_time-time) <= 1e-6

    def __step(self, time=0.0):
        r"""ODE IVP problem.
        Given the initial data at :math:`t=0`,
        :math:`u = (u_1(0),u_2(0),\ldots)`
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None
        """
        #import iapws.iapws97 as steam

        # Get state values
        u_0 = self.__get_state_vector(time)

        t_interval_sec = np.linspace(time, time+self.time_step, num=2)

        (u_vec_hist, info_dict) = odeint(self.__f_vec,
                                         u_0, t_interval_sec,
                                         args=(self.params,),
                                         rtol=1e-4, atol=1e-8, mxstep=200,
                                         full_output=True)

        assert info_dict['message'] == 'Integration successful.', info_dict['message']

        u_vec = u_vec_hist[1, :]  # solution vector at final time step

        n_dens = u_vec[0]
        c_vec = u_vec[1:7]
        fuel_temp = u_vec[7]
        cool_temp = u_vec[8]

        #update state variables
        outflow = self.coolant_outflow_phase.get_row(time)
        neutrons = self.neutron_phase.get_row(time)
        reactor = self.reactor_phase.get_row(time)
        self.params['outflow temp'] = cool_temp
        time += self.time_step

        self.coolant_outflow_phase.add_row(time, outflow)
        self.neutron_phase.add_row(time, neutrons)
        self.reactor_phase.add_row(time, reactor)

        self.coolant_outflow_phase.set_value('temp', cool_temp, time)
        self.neutron_phase.set_value('neutron-dens', n_dens, time)
        self.neutron_phase.set_value('delayed-neutrons-cc', c_vec, time)
        self.reactor_phase.set_value('fuel-temp', fuel_temp, time)

        return time

    def __get_signal_out(self, time):
        """Determine whether the RCIS module should be on or off.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        signal: string
            either 'online' or 'offline' to represent desired RCIS operation mode
        """
        if self.RCIS:
            signal = 'online'
            return(signal)
        else:
            signal = 'offline'
            return(signal)


    def __get_coolant_outflow(self, time):
        """Get the coolant outflow stream.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None

        """

        coolant_outflow_stream = dict()
        
        outflow_cool_temp = self.coolant_outflow_phase.get_value('temp', time)

        if outflow_cool_temp <= 373.15:
            coolant_outflow_stream['temp'] = outflow_cool_temp
            coolant_outflow_stream['pressure'] = 0
            coolant_outflow_stream['quality'] = 0
            coolant_outflow_stream['flowrate'] = 0

        else:

            if self.params['valve_opened']:
                pressure = steam._PSat_T(outflow_cool_temp)
                base_temp = 373.15
                base_pressure = steam._PSat_T(base_temp)
                base_params = steam.IAPWS97(T=base_temp, x=0)
                base_entropy = base_params.s
                base_cp = base_params.cp
                inlet_params = steam.IAPWS97(T=outflow_cool_temp, x=0)
                inlet_cp = inlet_params.cp
                average_cp = (base_cp + inlet_cp)/2

                delta_s = average_cp * math.log(outflow_cool_temp/base_temp) - (unit.R/1000) * math.log(pressure/base_pressure)

                inlet_entropy = (delta_s + base_entropy) * 1.25
                bubl_entropy = inlet_params.s
                bubl_enthalpy = inlet_params.h

                inlet_params = steam.IAPWS97(T=base_temp, x=1)
                dew_entropy = inlet_params.s
                dew_enthalpy = inlet_params.h
                quality = (inlet_entropy - bubl_entropy)/(dew_entropy - bubl_entropy)

                delta_t = time - self.params['valve_opening_time']
                mass_flowrate = 1820 * (1 - math.exp(-1 * delta_t/(3 * unit.minute)))

                coolant_outflow_stream['temp'] = outflow_cool_temp
                coolant_outflow_stream['pressure'] = pressure
                coolant_outflow_stream['quality'] = quality
                coolant_outflow_stream['flowrate'] = mass_flowrate

            else:
                self.params['valve_opened'] = True
                self.params['valve_opening_time'] = time
                coolant_outflow_stream['temp'] = 373.15
                coolant_outflow_stream['pressure'] = 0
                coolant_outflow_stream['quality'] = 0
                coolant_outflow_stream['flowrate'] = 0

        return coolant_outflow_stream

    def __get_state_vector(self, time):
        """Return a numpy array of all unknowns ordered as below:
            neutron density (1), delayed neutron emmiter concentrations (6),
            termperature of fuel (1), temperature of coolant (1).
        """

        u_vec = np.empty(0, dtype=np.float64)

        neutron_dens = self.neutron_phase.get_value('neutron-dens', time)
        u_vec = np.append(u_vec, neutron_dens)

        delayed_neutrons_cc = self.neutron_phase.get_value('delayed-neutrons-cc', time)
        u_vec = np.append(u_vec, delayed_neutrons_cc)

        fuel_temp = self.reactor_phase.get_value('fuel-temp', time)
        u_vec = np.append(u_vec, fuel_temp)

        temp = self.coolant_outflow_phase.get_value('temp', time)
        u_vec = np.append(u_vec, temp)

        return u_vec

    def __alpha_tn_func(self, temp, params):
        pressure = steam._PSat_T(temp) # vfda: why accessing protected method?

        d_rho = steam2.IAPWS95(P=pressure, T=temp-1).drhodT_P

        rho = 1 / steam._Region4(pressure, 0)['v'] # mass density, kg/m3

        n_m = ((rho * unit.kilo)/self.mod_molar_mass) * unit.N_A * (unit.centi)**3
        # number density of the moderator
        d_nm = ((d_rho * unit.kilo)/self.mod_molar_mass) * unit.N_A * (unit.centi)**3 #dNm/dTm
        d_nm = d_nm * unit.zepto * unit.milli

        mod_macro_a = self.mod_micro_a * n_m
        # macroscopic absorption cross section of the moderator
        mod_macro_s = self.mod_micro_s * n_m
        # macroscopic scattering cross section of the moderator

        F = self.fuel_macro_a/(self.fuel_macro_a + mod_macro_a) # thermal utilization, F
    #dF/dTm
        d_f = -1*(self.fuel_macro_a * self.mod_micro_a * d_nm)/(
            (self.fuel_macro_a + mod_macro_a)**2)

        # Resonance escape integral, P
        P = math.exp((-1 * self.n_fuel * (self.fuel_volume) * self.I)/
                     ((mod_macro_s * 3000)))
        #dP/dTm
        d_p = P * (-1 * self.n_fuel * self.fuel_volume * unit.centi**3 *
                   (self.mod_micro_s) * d_nm)/(mod_macro_s * 3000 * unit.centi**3)**2

        e_th = 0.0862 * temp # convert temperature to energy in MeV
        e_one = mod_macro_s/math.log(self.E0/e_th)
        # neutron thermalization macroscopic cross section

        d_f = 1/(3 * mod_macro_s * (1 - self.mod_mu0)) # neutron diffusion coefficient
        tau = d_f/e_one # fermi age, tau
        #dTau/dTm
        d_tau = (((0.0862 * (e_th/self.E0)) * 3 * n_m) - math.log(self.E0/e_th) *
                 ((self.mod_micro_s * d_nm))/((3 * n_m)**2 * (1 - self.mod_mu0)))

        L = math.sqrt(1/(3 * mod_macro_s * mod_macro_a * (1 - self.mod_mu0)))
        # diffusion length L
        # dL/dTm
        d_l = 1/(2 * math.sqrt((-2 * d_nm * unit.zepto * unit.milli)/(3 * self.mod_micro_s * (self.mod_micro_a * (n_m * unit.zepto * unit.milli)**3 * ((1 - self.mod_mu0))))))

        # left term of the numerator of the moderator temperature feedback coefficient, alpha
        left_first_term = d_tau * (self.buckling**2 + L**2 * self.buckling**4)
        #holding L as constant
        left_second_term = d_l * (2 * L * self.buckling**2 + 2 * L *
                                  (tau * self.buckling**4)) # holding tau as constant
        left_term = (P * F) * (left_first_term + left_second_term)
        # combining with P and F held as constant

        right_1st_term = (-1) * (1 + ((tau + d_l**2) * self.buckling**2) + \
                tau * d_l**2 * self.buckling**4) # num as const

        right_first_term = (-1) * ((1 + ((tau + L**2) * self.buckling**2))
                                   + tau * L**2 * self.buckling**4) # num as const
        right_second_term = F * d_p # holding thermal utilization as constant
        right_third_term = P * d_f # holding resonance escpae as constant
        right_term = right_first_term * (right_second_term + right_third_term)
        # combining all three terms together

        # numerator and denominator
        numerator = left_term + right_term
        denominator = self.eta * self.epsilon * \
                (F * P)**2

        alpha_tn = numerator/denominator


        alpha_tn = alpha_tn/3

        return alpha_tn

    def __rho_func(self, time, n_dens, temp, params):
        """Reactivity function.

        Parameters
        ----------
        t: float, required
            Time.
        temp_f: float, required
            Temperature at time t.
        params: dict, required
            Dictionary of quantities. It must have a `'rho_0'` key/value pair.

        Returns
        -------
        rho_t: float
            Value of reactivity.

        Examples
        --------
        """

        rho_0 = self.rho_0
        temp_ref = self.temp_o
        #n_dens_ss_operation = params['n_dens_ss_operation']
        alpha_n = self.alpha_n

        if temp < 293.15:
            # if temperature is less than the starting temperature then moderator
            # feedback is zero
            alpha_tn = 0

        else:
            alpha_tn = self.__alpha_tn_func(temp, self.params) #alpha_tn_func(temp, params)

        if params['malfunction start'] < time < params['malfunction end']:
            # reg rod held in position; only mod temp reactivity varies with time
            # during malfunction
            alpha_n = self.alpha_n_malfunction
            rho_t = rho_0 + alpha_n + alpha_tn * (temp - temp_ref)

        elif time > params['shutdown time']:
            # effectively the inverse of startup; gradually reduce reactivity and neutron density.
            rho_0 = -1 * rho_0
            alpha_n = rho_0 - (alpha_tn * (temp - temp_ref))
            rho_t = rho_0

        elif n_dens < 1e-5:
            #controlled startup w/ blade; gradually increase neutron density to SS value.
            #rho_current = (1 - n_dens) * rho_0
            #alpha_n = rho_current - rho_0 - alpha_tn * (temp - temp_ref)
            #rho_t = rho_current
            #self.alpha_n_malfunction = alpha_n
            rho_t = rho_0

        else:
            rho_current = (1 - n_dens) * rho_0
            alpha_n = rho_current - rho_0 - alpha_tn * (temp - temp_ref)
            rho_t = rho_current
            self.alpha_n_malfunction = alpha_n

        return (rho_t, alpha_n, alpha_tn * (temp - temp_ref))

    def __q_source(self, time, params):
        """Neutron source delta function.

        Parameters
        ----------
        t: float, required
            Time.
        params: dict, required
            Dictionary of quantities. It must have a `'q_0'` key/value pair.

        Returns
        -------
        q: float
            Value of source.

        Examples
        --------
        """
        q_0 = self.q_0

        if time <= 30 and params['shutdown-mode'] == False: # small time value
            q_source = q_0
        else:
            q_source = 0.0
            self.q_source_status = 'out'

        return q_source

    def __sigma_fis_func(self, temp, params):
        """Place holder for implementation
        """
        sigma_f = self.sigma_f_o  * math.sqrt(298/temp) * math.sqrt(math.pi) * 0.5

        return sigma_f

    def __nuclear_pwr_dens_func(self, time, temp, n_dens, params):
        """Place holder for implementation.
        """
        # include the neutrons from the initial source

        rxn_heat = self.fis_energy # get fission reaction energy J per reaction

        sigma_f = self.__sigma_fis_func(temp, self.params) # m2

        fis_nuclide_num_dens = self.fis_nuclide_num_dens #  #/m3

        sigma_fis = sigma_f * fis_nuclide_num_dens # macroscopic cross section

        v_o = self.thermal_neutron_velo # m/s

        neutron_flux = n_dens * 0.9E15 * v_o * sigma_f/self.sigma_f_o*0.7

         #reaction rate density
        rxn_rate_dens = sigma_fis * neutron_flux

        # nuclear power source
        q3prime = - rxn_heat * rxn_rate_dens # exothermic reaction W/m3)

        # add decay heat
        if self.params['shutdown-mode']:
            t_0 = unit.hour * 24 # assume 24 hours of steady state operation before shutdown
            p_0 = self.params['decay-heat-0']
            p_t = p_0 * (((time + 1) ** -0.2) - (t_0 + time) ** -0.2)
            q3prime += p_t
        
        return q3prime

    def __heat_sink_rate(self, time, temp_f, temp_c, params):

        ht_coeff = self.ht_coeff

        q_f = - ht_coeff * (temp_f - temp_c)
        return q_f

    def __f_vec(self, u_vec, time, params):
        num_negatives = u_vec[u_vec < 0]

        if num_negatives.any() < 0:
            assert np.max(abs(u_vec[u_vec < 0])) <= 1e-8, 'u_vec = %r'%u_vec

        #assert np.all(u_vec >= 0.0), 'u_vec = %r'%u_vec
        q_source_t = self.__q_source(time, self.params)

        n_dens = u_vec[0] # get neutron dens

        c_vec = u_vec[1:-2] # get delayed neutron emitter concentration

        temp_f = u_vec[-2] # get temperature of fuel

        temp_c = u_vec[-1] # get temperature of coolant

        # initialize f_vec to zero
        species_decay = self.species_decay
        lambda_vec = np.array(species_decay)
        n_species = len(lambda_vec)

        f_tmp = np.zeros(1+n_species+2, dtype=np.float64) # vector for f_vec return
        #----------------
        # neutron balance
        #----------------
        rho_t = self.__rho_func(time, n_dens, temp_c, self.params)[0]

        beta = self.beta
        gen_time = self.gen_time

        assert len(lambda_vec) == len(c_vec)

        f_tmp[0] = (rho_t - beta)/gen_time * n_dens + lambda_vec @ c_vec + q_source_t

        #-----------------------------------
        # n species balances (implicit loop)
        #-----------------------------------

        species_rel_yield = self.species_rel_yield
        beta_vec = np.array(species_rel_yield) * beta

        assert len(lambda_vec) == len(c_vec)
        assert len(beta_vec) == len(c_vec)

        f_tmp[1:-2] = beta_vec / gen_time * n_dens - lambda_vec * c_vec

        #--------------------
        # fuel energy balance
        #--------------------
        rho_f = self.fuel_dens
        cp_f = self.cp_fuel
        vol_fuel = self.fuel_volume

        pwr_dens = self.__nuclear_pwr_dens_func(time, (temp_f+temp_c)/2,
                                                n_dens, self.params)

        heat_sink = self.__heat_sink_rate(time, temp_f, temp_c, self.params)

        #assert heat_sink <= 0.0,'heat_sink = %r'%heat_sink

        f_tmp[-2] = -1/rho_f/cp_f * (pwr_dens - heat_sink/vol_fuel)

        #-----------------------
        # coolant energy balance
        #-----------------------
        rho_c = self.coolant_dens
        cp_c = self.cp_coolant
        vol_cool = self.coolant_volume

        # subcooled liquid
        pump_out = params['inflow-cool-temp']

        tau = self.tau

        heat_source = heat_sink
        temp_in = pump_out

        f_tmp[-1] = - 1/tau * (temp_c - temp_in) - 1./rho_c/cp_c/vol_cool * heat_source
        
        f_tmp[:] = [0 if np.isnan(x) else x for x in f_tmp] #nifty conditional mutator

        return f_tmp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

import os
import pickle
import logging
from cortix.src.port import Port

class Module:
    """Cortix module super class.

    This class provides facilities for creating modules within the Cortix network.
    Cortix will map one object of this class to either a Multiprocessing or MPI
    process depending on the user's configuration.

    Note
    ----
    This class is to be inherited by every Cortix module.
    In order to execute, modules *must* override the `run` method, which will be
    executed during the simulation.

    """

    def __init__(self):
        """Module super class constructor.

        Note
        ----
        This constructor must be called explicitly in the constructor of every
        Cortix module like so:

            super().__init__()

        Attributes
        ----------
        name: str
            A name given to the instance. Default is the derived class name.
        port_names_expected: list(str), None
            A list of names of ports expected in the module. This will be compared
            to port names during runtime to check against the intended use of the
            module.
        state: any
            Any `pickle-able` data structure to be passed in a `multiprocessing.Queue`
            to the parent process or to be gathered in the root MPI process.
            Default is `None`.
        use_mpi: bool
            `True` for MPI, `False` for Multiprocessing
        use_multiprocessing: bool
            `False` for MPI, `True` for Multiprocessing
        ports: list(Port)
            A list of ports contained by the module
        id: int
            An integer set by the external network once a module is added to it.
            The `id` is the position of the module in the network list.
            Default: None.
        __network: Network
            An internal network inherited by the derived module for nested networks.

       """

        self.name = self.__class__.__name__
        self.port_names_expected = None
        self.state = None
        self.use_mpi = False
        self.use_multiprocessing = True
        self.ports = list()
        self.log = logging.getLogger('cortix')
        self.save = False

        self.id = None

        self.__network = None

    def send(self, data, port):
        '''Send data through a given port.

        Parameters
        ----------
        data: any
            The data being sent out - must be pickleable
        port: Port, str
            A Port object to send the data through, or its string name

        '''

        if isinstance(port, str):
            port = self.get_port(port)
        elif isinstance(port, Port):
            assert port in self.ports, "Unknown port!"
        else:
            raise TypeError("port must be of Port or String type")

        port.send(data)

    def recv(self, port):
        '''Receive data from a given port

        Warning
        -------
        This function will block until data is available

        Parameters
        ----------
        port: Port, str
            A Port object to send the data through, or its string name

        Returns
        -------
        data: any
            The data received through the port

        '''

        if isinstance(port, str):
            port = self.get_port(port)
        elif isinstance(port, Port):
            assert port in self.ports, 'Unknown port!'
        else:
            raise TypeError('port must be of Port or String type')

        return port.recv()

    def get_port(self, name):
        '''Get port by name; if it does not exist, create one.

        Parameters
        ----------
        name: str
            The name of the port to get

        Returns
        -------
        port: Port
            The port object with the corresponding name

        '''

        assert isinstance(name, str), 'port name must be of type str'
        port = None

        for pti in self.ports:
            if pti.name == name:
                port = pti
                break

        if port is None:
            if self.port_names_expected:
                assert name in self.port_names_expected,\
                        'port name: {}, not expected by module: {}'.format(name, self)
            port = Port(name, self.use_mpi)
            self.ports.append(port)

        return port

    def __set_network(self, n):
        # Must be here to avoid infinite import loop
        from cortix.src.network import Network
        assert isinstance(n, Network)
        n.use_mpi = self.use_mpi
        n.use_multiprocessing = self.use_multiprocessing
        self.__network = n
    def __get_network(self):
        return self.__network
    network = property(__get_network, __set_network, None, None)

    def run(self, *args):
        '''Module run function

        Run method with an option to pass data back to the parent process when running
        in Python multiprocessing mode. If the user does not want to share data with
        the parent process, this function can be overriden with `run(self)`
        or `run(self, *args)` as long as `self.state = None`.
        If `self.state` points to anything but `None`, the user must use
        `run(self, *args).

        Notes
        -----
        When in multiprocessing, `*args` has two elements: `comm_idx` and `comm_state`.
        To pass back the state of the module, the user should insert the provided
        index `comm_idx` and the `state` into the queue as follows:

            if self.use_multiprocessing:
                try:
                    pickle.dumps(self.state)
                except pickle.PicklingError:
                    args[1].put((arg[0],None))
                else:
                    args[1].put((arg[0],self.state))

        at the bottom of the user defined `run()` function.

        Warning
        -------
        This function must be overridden by all Cortix modules

        Parameters
        ----------
        arg[0]: int
            Index of the state in the communication queue.

        arg[1]: multiprocessing.Queue
            When using the Python `multiprocessing` library `state_comm` must have
            the module's `self.state` in it. That is,
            `state_comm.put((idx_comm,self.state))` must be the last command in the
            method before `return`. In addition, self.state must be `pickle-able`.

        '''
        raise NotImplementedError('Module must implement run()')

    def run_and_save(self):

        self.run()

        if self.save:
            file_name = os.path.join('.ctx-saved', '{}_'.format(self.__class__.__name__))
            # Import where needed for no broken dependencies
            if self.use_mpi:
                from mpi4py import MPI
                file_name += str(MPI.COMM_WORLD.rank)
            else:
                file_name += str(os.getpid())
            file_name += '.pkl'

            self.ports = list() # reset ports since they can't be pickled
            self.log = None

            try:
                with open(file_name, 'wb') as fout:
                    pickle.dump(self, fout)
            except pickle.PicklingError:
                print('Unable to pickle {}!'.format(file_name))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

import logging

import numpy as np
import scipy.constants as const
from collections import namedtuple
import matplotlib
matplotlib.use('Agg', warn=False)
import matplotlib.pyplot as plt

from cortix.src.module import Module
from cortix.support.phase import Phase
from cortix.support.specie import Specie
from cortix.support.quantity import Quantity

class Vortex(Module):
    '''
    Vortex module used to model fluid flow using Cortix.

    Notes
    -----
    Any `port` name and any number of ports are allowed.

    '''

    def __init__(self):
        '''
        Attributes
        ----------
        initial_time: float
        end_time: float
        time_step: float
        show_time: tuple
            Two-element tuple, `(bool,float)`, `True` will print to standard
            output.

        '''

        super().__init__()

        species = list()
        quantities = list()

        self.initial_time = 0.0
        self.end_time = 5*const.minute
        self.time_step = 0.1
        self.show_time = (False,1*const.minute)
        self.log = logging.getLogger('cortix')

        air = Specie(name='air', formula_name='Air', phase='gas')
        air.massCCUnit = 'g/cc'
        air.molarCCUnit = 'mole/cc'
        air.molarMass = 0.3 * 16 * 2 + 0.7 * 14 *2
        species.append(air)

        # Constant values for the vortex fluid.
        self.mass_density  = 0.1 * const.gram / const.centi**3 # [kg/m^3]
        self.dyn_viscosity = 1.81e-5 # kg/(m s)

        # Domain box dimensions: LxLxH m^3 box with given H.
        # z coordinate pointing upwards. -L <= x <= L, -L <= y <= L, 
        # z component is positive => vortex is blowing upwards.
        self.box_half_length = 250.0 # vortex box [m] 
        self.box_height      = 500.0 # [m]

        # Vortex parameters.
        self.min_core_radius = 2.5 # [m]
        self.outer_v_theta   = 1.0 # m/s # angular speed
        self.v_z_0 = 0.50 # [m/s]

        self.period = 20 # wind change period

    def run(self, *args):

        # namedtuple does not pickle into send message; investigate later: vfda TODO
        #Props = namedtuple('Props',['mass_density','dyn_viscosity'])
        #fluid_props = Props( self.mass_density, self.dyn_viscosity )
        fluid_props = ( self.mass_density, self.dyn_viscosity )

        time = self.initial_time

        while time < self.end_time:

            if self.show_time[0] and abs(time%self.show_time[1]-0.0)<=1.e-1:
                self.log.info('Vortex::time[min] = '+str(round(time/const.minute,1)))

            # Interactions in all nameless ports (lower level port send/recv used)
            #---------------------------------------------------------------------

            for port in self.ports:
                (message_time, position) = port.recv()

                # Compute the vortex velocity using the given position
                velocity = self.compute_velocity(message_time, position)

                # Send the vortex velocity to caller
                port.send( (message_time, velocity, fluid_props) )

            time += self.time_step

        return

    def compute_velocity(self, time, position):
        '''
        Compute the vortex velocity at the given external
        position using a vortex flow model

        Parameters
        ----------
        time: float
            Time in SI unit.
        position: numpy.ndarray(3)
            Spatial position in SI unit.

        Returns
        -------
        vortex_velocity: numpy.ndarray(3)

        '''
        import math

        cycle_freq  = 1/self.period
        radian_freq = 2*math.pi*cycle_freq

        outer_cylindrical_radius = np.hypot(self.box_half_length, self.box_half_length)
        circulation = 2 * np.pi * outer_cylindrical_radius * self.outer_v_theta # m^2/s
        core_radius = self.min_core_radius

        x = position[0]
        y = position[1]
        z = position[2]

        relax_length = self.box_height / 2.0
        z_relax_factor = np.exp(-(self.box_height-z)/relax_length)
        v_z = self.v_z_0 * z_relax_factor * abs(math.cos( radian_freq * time))

        cylindrical_radius = np.hypot(x,y)
        azimuth = np.arctan2(y,x)

        v_theta = (1 - np.exp(-cylindrical_radius**2 / 8 / core_radius**2)) *\
                   circulation / 2 / np.pi /\
                   max(cylindrical_radius,self.min_core_radius) *\
                   z_relax_factor * abs(math.cos( radian_freq * time))

        v_x = - v_theta * np.sin(azimuth)
        v_y =   v_theta * np.cos(azimuth)

        return np.array([v_x,v_y,v_z])

    def plot_velocity(self, time=None):
        '''
        Plot the vortex velocity as a function of height.

        '''

        if time is None:
            time = self.initial_time

        (fig,axs) = plt.subplots(2,2)
        fig.subplots_adjust(hspace=0.5, wspace=0.5)

        for z in np.flip(np.linspace(0, self.box_height,3), 0):
            xval = list()
            yval = list()
            for x in np.linspace(0, self.box_half_length, 500):
                xval.append(x)
                y = 0.0
                vortex_velocity = self.compute_velocity( time,np.array([x,y,z]) )
                yval.append(vortex_velocity[1])

            axs[0,0].plot(xval, yval, label='z =' + str(round(z,2))+' [m]')

        axs[0,0].set_xlabel('Radial distance [m]')
        axs[0,0].set_ylabel('Tangential speed [m/s]')
        axs[0,0].legend(loc='best')
        fig.suptitle('Vortex Flow at '+str(round(time,1))+' [s]')
        axs[0,0].grid(True)

        xval = list()
        yval = list()
        for z in np.linspace(0,self.box_height,50):
            yval.append(z)
            vortex_velocity = self.compute_velocity( time,np.array([0.0,0.0,z]) )
            xval.append(vortex_velocity[2])

        axs[0,1].plot(xval,yval)
        axs[0,1].set_xlabel('Vertical speed [m/s]')
        axs[0,1].set_ylabel('Height [m]')
        axs[0,1].grid(True)

        fig.savefig('vortex_velocity.png',dpi=200)
        plt.close(fig)

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
'''
UMass Lowell research nuclear reactor single point dynamics.
'''

import scipy.constants as unit
import matplotlib.pyplot as plt

from cortix import Cortix
from cortix import Network

from umlrr import UMLRR

def main():
    '''Cortix run file for UMass Lowell research nuclear reactor.

    Attributes
    ----------
    end_time: float
        End of the flow time in SI unit.
    time_step: float
        Size of the time step between port communications in SI unit.
    use_mpi: bool
        If set to `True` use MPI otherwise use Python multiprocessing.

    '''

    # Preamble 

    end_time  = 1.0 * unit.hour
    time_step = 0.5 * unit.minute
    show_time = (True,15*unit.minute)

    use_mpi = False  # True for MPI; False for Python multiprocessing

    # System top level
    radlab = Cortix( use_mpi=use_mpi, splash=True )

    # Network
    radlab_net = radlab.network = Network()

    #from shutdown_params import shutdown_params
    #params = shutdown_params()

    # Create reactor module
    reactor = UMLRR()

    reactor.name      = 'UMLRR'
    reactor.save      = True
    reactor.time_step = time_step
    reactor.end_time  = end_time
    reactor.show_time = show_time

    # Add reactor module to network
    radlab_net.module(reactor)

    # Create the network connectivity
    #radlab_net.connect( [reactor, 'none'], [none,'none'] )

    #*****************************************************************************

    radlab_net.draw()

    # Run network dynamics simulation
    radlab.run()

    plot_results = True

    if plot_results and ( radlab.use_multiprocessing or radlab.rank == 0 ):

        # Reactor graphs
        reactor = radlab_net.modules[0]

        (quant, time_unit) = reactor.reactor_phase.get_quantity_history('neutron-dens')

        quant.plot( x_scaling=1/unit.minute, x_label='Time [m]',
                    y_label=quant.formal_name+' ['+quant.unit+']')
        plt.grid()
        plt.savefig('neutron-dens.png', dpi=300)

        (quant, time_unit) = reactor.reactor_phase.get_quantity_history('delayed-neutrons-cc')
        quant.plot( x_scaling=1/unit.minute, x_label='Time [m]',
                    y_label=quant.formal_name+' ['+quant.unit+']')
        plt.grid()
        plt.savefig('delayed-neutrons-cc.png', dpi=300)

        (quant, time_unit) = reactor.reactor_phase.get_quantity_history('fuel-temp')

        quant.plot( x_scaling=1/unit.minute, x_label='Time [m]',
                    y_label =quant.formal_name+' ['+quant.unit+']')
        plt.grid()
        plt.savefig('fuel-temp.png', dpi=300)

        (quant, time_unit) = reactor.reactor_phase.get_quantity_history('power')

        quant.plot( x_scaling=1/unit.minute, x_label='Time [m]',
                    y_scaling=1/unit.kilo,
                    y_label=quant.formal_name+' ['+'k'+quant.unit+']')
        plt.grid()
        plt.savefig('power.png', dpi=300)

        (quant, time_unit) = reactor.coolant_phase.get_quantity_history('temp')

        quant.plot( x_scaling=1/unit.minute, x_label='Time [m]',
                    y_label =quant.formal_name+' ['+quant.unit+']')
        plt.grid()
        plt.savefig('cool-temp.png', dpi=300)

    # Properly shutdow radlab
    radlab.close()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org

import logging

import numpy as np
import scipy.constants as unit
from scipy.integrate import odeint
import math

from cortix import Module
from cortix.support.phase_new import PhaseNew as Phase
from cortix import Quantity

class BWR(Module):
    '''
    Boiling water reactor single-point reactor.

    Notes
    -----
    These are the `port` names available in this module to connect to respective
    modules: `turbine`, and `pump`.
    See instance attribute `port_names_expected`.

    '''

    def __init__(self, params):
        '''
        Parameters
        ----------
        params: dict
            All parameters for the module in the form of a dictionary.

        '''

        super().__init__()

        self.port_names_expected = ['coolant-inflow', 'coolant-outflow',
                                     'signal-out', 'signal-in']

        self.params = params

        self.initial_time = 0.0 * unit.day
        self.end_time     = 4 * unit.hour
        self.time_step    = 10.0

        self.show_time    = (False,10.0)

        self.log = logging.getLogger('cortix')

        # Coolant outflow phase history
        quantities = list()

        flowrate = Quantity(name='flowrate', formal_name='q_c', unit='kg/s', value=0.0,
                info='Reactor Outflow Coolant Flowrate' ,latex_name='$q_c$')

        quantities.append(flowrate)

        temp = Quantity(name='temp', formal_name='T_c', unit='K', value=273.15,
                info='Reactor Outflow Coolant Temperature', latex_name='$T_c$')

        quantities.append(temp)

        press = Quantity(name='pressure', formal_name='P_c', unit='Pa', value=0.0,
                info='Reactor Outflow Coolant Pressure', latex_name='$P_c$')

        quantities.append(press)

        quality = Quantity(name='steam-quality', formal_name='chi_s', unit='', value=0.0,
                info='Reactor STeam Quality', latex_name='$\chi$')

        quantities.append(quality)

        self.coolant_outflow_phase = Phase(self.initial_time, time_unit='s',
                quantities=quantities)

        # Neutron phase history
        quantities = list()

        neutron_dens = Quantity(name='neutron-dens', formal_name='n', unit='1/m^3',
                value=0.0, info='Rel. Reactor Neutron Density', latex_name='$n$')

        quantities.append(neutron_dens)

        delayed_neutrons_0 = np.zeros(6)

        delayed_neutron_cc = Quantity(name='delayed-neutrons-cc', formal_name='c_i',
                unit='1/m^3 ', value=delayed_neutrons_0,
                info='Rel. Delayed Neutron Precursors', latex_name='$c_i$')

        quantities.append(delayed_neutron_cc)

        self.neutron_phase = Phase(self.initial_time, time_unit='s',
                quantities=quantities)

        #reactor paramaters
        quantities = list()

        fuel_temp = Quantity( name='fuel-temp', formalName='T_f', unit='k',
                value=273.15, info='Reactor Nuclear Fuel Temperature',
                latex_name='$T_f$')

        quantities.append(fuel_temp)

        reg_rod_position = Quantity(name='reg-rod-position',
                formal_name = 'reg rod position', unit='m', value=0.0,
                info='Reactor Regulating Rod Position', latex_name='$x_p$')

        quantities.append(reg_rod_position)

        self.reactor_phase = Phase(self.initial_time, time_unit='s',
                quantities=quantities)

        # Initialize inflow
        self.params['inflow-cool-temp'] = 273.15

        return

    def run(self, *args):

        # Some logic for logging time stamps
        if self.initial_time + self.time_step > self.end_time:
            self.end_time = self.initial_time + self.time_step

        time = self.initial_time
        print_time = self.initial_time
        print_time_step = self.show_time[1]
        if print_time_step < self.time_step:
            print_time_step = self.time_step

        while time <= self.end_time:

            last_time_stamp = time - self.time_step

            if self.show_time[0] and time>=print_time and \
                    time<print_time+print_time_step:

                self.log.info( self.name+'::run():time[m]='+
                        str(round(time/unit.minute,1)))

                self.__logit = True
                print_time += self.show_time[1]

            else:
                self.__logit = False

            # Communicate information
            #------------------------
            self.__call_ports(time)

            # Evolve one time step
            #---------------------

            time = self.__step( time )

    def __call_ports(self, time):

        # Interactions in the coolant-outflow port
        #-----------------------------------------
        # one way "to" coolant-outflow

        # send to 
        if self.get_port('coolant-outflow').connected_port:

            message_time = self.recv('coolant-outflow')

            coolant_outflow = self.__get_coolant_outflow( message_time )
            outflow_params = dict()
            self.send( (message_time, coolant_outflow), 'coolant-outflow' )

        # Interactions in the coolant-inflow port
        #----------------------------------------
        # one way "from" coolant-inflow

        #self.send( time, 'coolant-inflow' )

        # receive from
        if self.get_port('coolant-inflow').connected_port:

            self.send( time, 'coolant-inflow' )
            (check_time, inflow_cool_temp) = self.recv('coolant-inflow')

            assert abs(check_time-time) <= 1e-6
            self.params['inflow-cool-temp'] = inflow_cool_temp

        # Interactions in the signal-out port
        #-----------------------------------------
        # one way "to" signal-out 

        # send to 
        if self.get_port('signal-out').connected_port:

            message_time = self.recv('signal-out')

            signal_out = self.__get_signal_out(time)

            self.send( (message_time, signal_out), 'signal-out' )

    def __get_coolant_outflow(message_time):

        outflow = self.params['coolant-outflow']
        return(outflow)

    def __step(self, time=0.0):
        r'''
        ODE IVP problem:
        Given the initial data at :math:`t=0`,
        :math:`u = (u_1(0),u_2(0),\ldots)`
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None

        '''
        import iapws.iapws97 as steam

        # Get state values
        u_0 = self.__get_state_vector( time )

        t_interval_sec = np.linspace(time, time+self.time_step, num=2)

        (u_vec_hist, info_dict) = odeint( self.__f_vec,
                                          u_0, t_interval_sec,
                                          args=( self.params, ),
                                          rtol=1e-4, atol=1e-8, mxstep=200,
                                          full_output=True )

        assert info_dict['message'] =='Integration successful.', info_dict['message']

        u_vec = u_vec_hist[1,:]  # solution vector at final time step

        n_dens    = u_vec[0]
        c_vec     = u_vec[1:7]
        fuel_temp = u_vec[7]
        cool_temp = u_vec[8]

        #update state variables
        outflow  = self.coolant_outflow_phase.get_row(time)
        neutrons = self.neutron_phase.get_row(time)
        reactor  = self.reactor_phase.get_row(time)
        self.params['outflow temp'] = cool_temp
        time += self.time_step

        self.coolant_outflow_phase.add_row(time, outflow)
        self.neutron_phase.add_row(time, neutrons)
        self.reactor_phase.add_row(time, reactor)

        self.coolant_outflow_phase.set_value('temp', cool_temp, time)
        self.neutron_phase.set_value('neutron-dens', n_dens, time)
        self.neutron_phase.set_value('delayed-neutrons-cc', c_vec, time)
        self.reactor_phase.set_value('fuel-temp', fuel_temp, time)

        return time

    def __signal_out(self, time=0.0):
        '''
        Send reactor signals to anyone requesting it.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None

        '''

        signals = dict()

        return signals

    def __get_coolant_outflow(self, time=0.0):
        '''
        Get the coolant outflow stream.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None

        '''

        coolant_outflow_stream = dict()

        outflow_cool_temp = self.coolant_outflow_phase.get_value('temp', time)

        coolant_outflow_stream['temp'] = outflow_cool_temp
        coolant_outflow_stream['quality'] = 0.7
        return coolant_outflow_stream

    def __get_state_vector(self, time):
        '''
        Return a numpy array of all unknowns ordered as below:
            neutron density (1), delayed neutron emmiter concentrations (6),
            termperature of fuel (1), temperature of coolant (1).
        '''

        u_vec = np.empty(0,dtype=np.float64)

        neutron_dens = self.neutron_phase.get_value('neutron-dens',time)
        u_vec = np.append( u_vec, neutron_dens )

        delayed_neutrons_cc = self.neutron_phase.get_value('delayed-neutrons-cc',time)
        u_vec = np.append(u_vec, delayed_neutrons_cc)

        fuel_temp = self.reactor_phase.get_value('fuel-temp',time)
        u_vec = np.append( u_vec, fuel_temp)

        temp = self.coolant_outflow_phase.get_value('temp',time)
        u_vec = np.append(u_vec, temp)

        # sanity check

        return u_vec

    def __alpha_tn_func(self, temp, params):
        import math
        import scipy.misc as diff
        import scipy.constants as unit
        import iapws.iapws97 as steam
        import iapws.iapws95 as steam2

        pressure = steam._PSat_T(temp)

        d_rho = steam2.IAPWS95(P=pressure, T=temp-1).drhodT_P

        #d_rho2 = diff.derivative(derivative_helper, temp) # dRho/dTm

        rho = 1 / steam._Region4(pressure, 0)['v'] # mass density, kg/m3

        Nm = ((rho * unit.kilo)/params['mod molar mass']) * unit.N_A * (unit.centi)**3 # number density of the moderator
        d_Nm =  ((d_rho * unit.kilo)/params['mod molar mass']) * unit.N_A * (unit.centi)**3 #dNm/dTm
        d_Nm = d_Nm * unit.zepto * unit.milli

        mod_macro_a = params['mod micro a'] * Nm # macroscopic absorption cross section of the moderator
        mod_macro_s = params['mod micro s'] * Nm # macroscopic scattering cross section of the moderator

        F = params['fuel macro a']/(params['fuel macro a'] + mod_macro_a) # thermal utilization, F
    #dF/dTm
        d_F = -1*(params['fuel macro a'] * params['mod micro a'] * d_Nm)/(params['fuel macro a'] + mod_macro_a)**2

        # Resonance escape integral, P
        P = math.exp((-1 * params['n fuel'] * (params['fuel_volume']) * params['I'])/(mod_macro_s * 3000))
        #dP/dTm
        d_P = P * (-1 * params['n fuel'] * params['fuel_volume'] * unit.centi**3 * params['mod micro s'] * d_Nm)/(mod_macro_s * 3000 * unit.centi**3)**2

        Eth = 0.0862 * temp # convert temperature to energy in MeV
        E1 = mod_macro_s/math.log(params['E0']/Eth) # neutron thermalization macroscopic cross section

        Df = 1/(3 * mod_macro_s * (1 - params['mod mu0'])) # neutron diffusion coefficient
        tau = Df/E1 # fermi age, tau
        #dTau/dTm
        d_tau = (((0.0862 * (Eth/params['E0'])) * 3 * Nm) - math.log(params['E0']/Eth) * (params['mod micro s'] * d_Nm))/((3 * Nm)**2 * (1 - params['mod mu0']))

        L = math.sqrt(1/(3 * mod_macro_s * mod_macro_a * (1 - params['mod mu0']))) # diffusion length L
        # dL/dTm
        d_L = 1/(2 * math.sqrt((-2 * d_Nm * unit.zepto * unit.milli)/(3 * params['mod micro s'] * params['mod micro a'] * (Nm * unit.zepto * unit.milli)**3 * (1 - params['mod mu0']))))

        # left term of the numerator of the moderator temperature feedback coefficient, alpha
        left_1st_term = d_tau * (params['buckling']**2 + L**2 * params['buckling']**4) #holding L as constant
        left_2nd_term = d_L * (2 * L * params['buckling']**2 + 2 * L * tau * params['buckling']**4) # holding tau as constant
        left_term = (P * F) * (left_1st_term + left_2nd_term) # combining with P and F held as constant

        # right term of the numerator of the moderator temperature feedback coefficient, alpha

        right_1st_term = (-1) * (1 + ((tau + L**2) * params['buckling']**2) + tau * L**2 * params['buckling']**4) # num as const
        right_2nd_term = F * d_P # holding thermal utilization as constant
        right_3rd_term = P * d_F # holding resonance escpae as constant
        right_term = right_1st_term * (right_2nd_term + right_3rd_term) # combining all three terms together

        # numerator and denominator
        numerator = left_term + right_term
        denominator = params['eta'] * params['epsilon'] * (F * P)**2

        alpha_tn = numerator/denominator


        alpha_tn = alpha_tn/3
        return alpha_tn

    def __rho_func(self, time, n_dens, temp, params, ):
        '''
        Reactivity function.

        Parameters
        ----------
        t: float, required
            Time.
        temp_f: float, required
            Temperature at time t.
        params: dict, required
            Dictionary of quantities. It must have a `'rho_0'` key/value pair.

        Returns
        -------
        rho_t: float
            Value of reactivity.

        Examples
        --------
        '''

        rho_0    = params['rho_0']
        temp_ref = params['temp_0']
        n_dens_ss_operation = params['n_dens_ss_operation']
        alpha_n  = params['alpha_n']

        if temp < 293.15: # if temperature is less than the starting temperature then moderator feedback is zero
            alpha_tn = 0

        else:
            alpha_tn = self.__alpha_tn_func(temp , self.params) #alpha_tn_func(temp, params)

        if time > params['malfunction start'] and time < params['malfunction end']: # reg rod held in position; only mod temp reactivity varies with time during malfunction
            alpha_n = params['alpha_n_malfunction']
            rho_t = rho_0 + alpha_n + alpha_tn * (temp - temp_ref)

        elif time > params['shutdown time']: # effectively the inverse of startup; gradually reduce reactivity and neutron density.
            rho_0 = -1 * rho_0
            alpha_n = rho_0 - (alpha_tn * (temp - temp_ref))
            rho_t = rho_0

        elif n_dens < 1e-5: #controlled startup w/ blade; gradually increase neutron density to SS value.
            #rho_current = (1 - n_dens) * rho_0
            #alpha_n = rho_current - rho_0 - alpha_tn * (temp - temp_ref)
            #rho_t = rho_current
            #params['alpha_n_malfunction'] = alpha_n
            rho_t = rho_0

        else:
            rho_current = (1 - n_dens) * rho_0
            alphh_n = rho_current - rho_0 - alpha_tn * (temp - temp_ref)
            rho_t = rho_current
            params['alpha_n_malfunction'] = alpha_n
        #print(n_dens)

        return (rho_t, alpha_n, alpha_tn * (temp - temp_ref))

    def __q_source(self, time, params):
        '''
        Neutron source delta function.

        Parameters
        ----------
        t: float, required
            Time.
        params: dict, required
            Dictionary of quantities. It must have a `'q_0'` key/value pair.

        Returns
        -------
        q: float
            Value of source.

        Examples
        --------
        '''
        q_0 = params['q_0']

        if time <= 1500: # small time value
            q = q_0
        else:
            q = 0.0
            params['q_source_status'] = 'out'

        return q

    def __sigma_fis_func(self, temp, params):
        '''
        Place holder for implementation
        '''
        sigma_f = params['sigma_f_o']  * math.sqrt(298/temp) * math.sqrt(math.pi) * 0.5

        return(sigma_f)

    def __nuclear_pwr_dens_func(self, time, temp, n_dens, params ):
        '''
        Place holder for implementation
        '''
        n_dens = n_dens + self.__q_source(time, self.params) # include the neutrons from the initial source

        rxn_heat = params['fis_energy'] # get fission reaction energy J per reaction

        sigma_f = self.__sigma_fis_func( temp, self.params ) # m2

        fis_nuclide_num_dens = params['fis_nuclide_num_dens_fake'] #  #/m3

        Sigma_fis = sigma_f * fis_nuclide_num_dens # macroscopic cross section

        v_o = params['thermal_neutron_velo'] # m/s

        neutron_flux = n_dens * 0.95E15 * v_o * sigma_f/params['sigma_f_o']*0.7

         #reaction rate density
        rxn_rate_dens = Sigma_fis * neutron_flux

        # nuclear power source
        q3prime = - rxn_heat * rxn_rate_dens # exothermic reaction W/m3)
        #q3prime = - n_dens * 3323E6
        #print("q3prime")
        #print(q3prime)

        return q3prime

    def __heat_sink_rate(self, time, temp_f, temp_c, params):

        ht_coeff = params['ht_coeff']

        q_f = - ht_coeff * (temp_f - temp_c)
        #print(q_f)
        return q_f

    def __f_vec(self, u_vec, time, params):
        num_negatives = u_vec[u_vec < 0]

        if num_negatives.any() < 0:
            assert np.max(abs(u_vec[u_vec < 0])) <= 1e-8, 'u_vec = %r'%u_vec

        #assert np.all(u_vec >= 0.0), 'u_vec = %r'%u_vec
        q_source_t = self.__q_source(time, self.params)

        n_dens = u_vec[0] # get neutron dens

        c_vec  = u_vec[1:-2] # get delayed neutron emitter concentration

        temp_f = u_vec[-2] # get temperature of fuel

        temp_c = u_vec[-1] # get temperature of coolant

        # initialize f_vec to zero 
        species_decay = params['species_decay']
        lambda_vec = np.array(species_decay)
        n_species  = len(lambda_vec)

        f_tmp = np.zeros(1+n_species+2,dtype=np.float64) # vector for f_vec return

        #----------------
        # neutron balance
        #----------------
        rho_t = self.__rho_func(time, n_dens, temp_c, self.params)[0]

        beta = params['beta']
        gen_time = params['gen_time']

        assert len(lambda_vec)==len(c_vec)

        f_tmp[0] = (rho_t - beta)/gen_time * n_dens + lambda_vec @ c_vec + q_source_t

        #-----------------------------------
        # n species balances (implicit loop)
        #-----------------------------------

        species_rel_yield = params['species_rel_yield']
        beta_vec = np.array(species_rel_yield) * beta

        assert len(lambda_vec)==len(c_vec)
        assert len(beta_vec)==len(c_vec)
        #species are in RELATIVE yield, not actual; multiply by delayed neutron
        #yield fraction of 0.0065 to get the actual delayed neutron
        #concentration

        f_tmp[1:-2] = beta_vec / gen_time * n_dens - lambda_vec * c_vec


        #--------------------
        # fuel energy balance
        #--------------------
        rho_f    = params['fuel_dens']
        cp_f     = params['cp_fuel']
        vol_fuel = params['fuel_volume']

        pwr_dens  = self.__nuclear_pwr_dens_func( time, (temp_f+temp_c)/2, n_dens, self.params)

        heat_sink = self.__heat_sink_rate( time, temp_f, temp_c, self.params)

        #assert heat_sink <= 0.0,'heat_sink = %r'%heat_sink

        f_tmp[-2] =  -1/rho_f/cp_f * ( pwr_dens - heat_sink/vol_fuel )

        #-----------------------
        # coolant energy balance
        #-----------------------
        rho_c    = params['coolant_dens']
        cp_c     = params['cp_coolant']
        vol_cool = params['coolant_volume']

        # subcooled liquid
        pump_out = params['inflow-cool-temp']

        tau = params['tau_fake']

        heat_source = heat_sink
        temp_in = pump_out

        f_tmp[-1] = - 1/tau * (temp_c - temp_in) - 1./rho_c/cp_c/vol_cool * heat_source

        # pressure calculations

        #print(time)
        #print(u_vec)
        return f_tmp
